"""Useful functions for working with vectors and matrices."""

import numpy as np
import scipy.sparse as sp
from discretize.utils.code_utils import is_scalar, deprecate_function
import warnings


def mkvc(x, n_dims=1, **kwargs):
    """Coerce a vector to the specified dimensionality.

    This function converts a :class:`numpy.ndarray` to a vector. In general,
    the output vector has a dimension of 1. However, the dimensionality
    can be specified if the user intends to carry out a dot product with
    a higher order array.

    Parameters
    ----------
    x : array_like
        An array that will be reorganized and output as a vector. The input array
        will be flattened on input in Fortran order.
    n_dims : int
        The dimension of the output vector. :data:`numpy.newaxis` are appended to the
        output array until it has this many axes.

    Returns
    -------
    numpy.ndarray
        The output vector, with at least ``n_dims`` axes.

    Examples
    --------
    Here, we reorganize a simple 2D array as a vector and demonstrate the
    impact of the *n_dim* argument.

    >>> from discretize.utils import mkvc
    >>> import numpy as np
    >>> rng = np.random.default_rng(856)

    >>> a = rng.random(3, 2)
    >>> a
    array([[0.33534155, 0.25334363],
           [0.07147884, 0.81080958],
           [0.85892774, 0.74357806]])

    >>> v = mkvc(a)
    >>> v
    array([0.33534155, 0.07147884, 0.85892774, 0.25334363, 0.81080958,
           0.74357806])

    In Higher dimensions:

    >>> for ii in range(1, 4):
    ...     v = mkvc(a, ii)
    ...     print('Shape of output with n_dim =', ii, ': ', v.shape)
    Shape of output with n_dim = 1 :  (6,)
    Shape of output with n_dim = 2 :  (6, 1)
    Shape of output with n_dim = 3 :  (6, 1, 1)
    """
    if "numDims" in kwargs:
        raise TypeError(
            "The numDims keyword argument has been removed, please use n_dims. "
            "This will be removed in discretize 1.0.0",
        )
    if isinstance(x, np.matrix):
        x = np.array(x)

    if hasattr(x, "tovec"):
        x = x.tovec()

    if isinstance(x, Zero):
        return x

    if not isinstance(x, np.ndarray):
        raise TypeError("Vector must be a numpy array")

    if n_dims == 1:
        return x.flatten(order="F")
    elif n_dims == 2:
        return x.flatten(order="F")[:, np.newaxis]
    elif n_dims == 3:
        return x.flatten(order="F")[:, np.newaxis, np.newaxis]


def sdiag(v):
    """Generate sparse diagonal matrix from a vector.

    This function creates a sparse diagonal matrix whose diagonal elements
    are defined by the input vector *v*. For a vector of length *n*,
    the output matrix has shape (n,n).

    Parameters
    ----------
    v : (n) numpy.ndarray or discretize.utils.Zero
        The vector defining the diagonal elements of the sparse matrix being constructed

    Returns
    -------
    (n, n) scipy.sparse.csr_matrix or discretize.utils.Zero
        The sparse diagonal matrix.

    Examples
    --------
    Use a 1D array of values to construct a sparse diagonal matrix.

    >>> from discretize.utils import sdiag
    >>> import numpy as np
    >>> v = np.array([6., 3., 1., 8., 0., 5.])
    >>> M = sdiag(v)
    """
    if isinstance(v, Zero):
        return Zero()

    return sp.spdiags(mkvc(v), 0, v.size, v.size, format="csr")


def sdinv(M):
    """Return inverse of a sparse diagonal matrix.

    This function extracts the diagonal elements of the input matrix *M*
    and creates a sparse diagonal matrix from the reciprocal these elements.
    If the input matrix *M* is diagonal, the output is the inverse of *M*.

    Parameters
    ----------
    M : (n, n) scipy.sparse.csr_matrix
        A sparse diagonal matrix

    Returns
    -------
    (n, n) scipy.sparse.csr_matrix
        The inverse of the sparse diagonal matrix.

    Examples
    --------
    >>> from discretize.utils import sdiag, sdinv
    >>> import numpy as np

    >>> v = np.array([6., 3., 1., 8., 0., 5.])
    >>> M = sdiag(v)
    >>> Minv = sdinv(M)

    """
    return sdiag(1.0 / M.diagonal())


def speye(n):
    """Generate sparse identity matrix.

    Parameters
    ----------
    n : int
        The dimensions of the sparse identity matrix.

    Returns
    -------
    (n, n) scipy.sparse.csr_matrix
        The sparse identity matrix.
    """
    return sp.identity(n, format="csr")


def kron3(A, B, C):
    r"""Compute kronecker products between 3 sparse matricies.

    Where :math:`\otimes` denotes the Kronecker product and *A, B* and *C* are
    sparse matrices, this function outputs :math:`(A \otimes B) \otimes C`.

    Parameters
    ----------
    A, B, C : scipy.sparse.spmatrix
        Sparse matrices.

    Returns
    -------
    scipy.sparse.csr_matrix
        Kroneker between the 3 sparse matrices.
    """
    return sp.kron(sp.kron(A, B), C, format="csr")


def spzeros(n1, n2):
    """Generate sparse matrix of zeros of shape=(n1, n2).

    Parameters
    ----------
    n1 : int
        Number of rows.
    n2 : int
        Number of columns.

    Returns
    -------
    (n1, n2) scipy.sparse.dia_matrix
        A sparse matrix of zeros.
    """
    return sp.dia_matrix((n1, n2))


def ddx(n):
    r"""Create 1D difference (derivative) operator from nodes to centers.

    For n cells, the 1D difference (derivative) operator from nodes to
    centers is sparse, has shape (n, n+1) and takes the form:

    .. math::
        \begin{bmatrix}
        -1 & 1 & & & \\
        & -1 & 1 & & \\
        & & \ddots & \ddots & \\
        & & & -1 & 1
        \end{bmatrix}

    Parameters
    ----------
    n : int
        Number of cells

    Returns
    -------
    (n, n + 1) scipy.sparse.csr_matrix
        The 1D difference operator from nodes to centers.
    """
    return sp.spdiags((np.ones((n + 1, 1)) * [-1, 1]).T, [0, 1], n, n + 1, format="csr")


def av(n):
    r"""Create 1D averaging operator from nodes to cell-centers.

    For n cells, the 1D averaging operator from nodes to centerss
    is sparse, has shape (n, n+1) and takes the form:

    .. math::
        \begin{bmatrix}
        1/2 & 1/2 & & & \\
        & 1/2 & 1/2 & & \\
        & & \ddots & \ddots & \\
        & & & 1/2 & 1/2
        \end{bmatrix}

    Parameters
    ----------
    n : int
        Number of cells

    Returns
    -------
    (n, n + 1) scipy.sparse.csr_matrix
        The 1D averaging operator from nodes to centers.
    """
    return sp.spdiags(
        (0.5 * np.ones((n + 1, 1)) * [1, 1]).T, [0, 1], n, n + 1, format="csr"
    )


def av_extrap(n):
    r"""Create 1D averaging operator from cell-centers to nodes.

    For n cells, the 1D averaging operator from cell centers to nodes
    is sparse and has shape (n+1, n). Values at the outmost nodes are
    extrapolated from the nearest cell center value. Thus the operator
    takes the form:

    .. math::
        \begin{bmatrix}
        1 & & & & \\
        1/2 & 1/2 & & & \\
        & 1/2 & 1/2 & & & \\
        & & \ddots & \ddots & \\
        & & & 1/2 & 1/2 \\
        & & & & 1
        \end{bmatrix}

    Parameters
    ----------
    n : int
        Number of cells

    Returns
    -------
    (n+1, n) scipy.sparse.csr_matrix
        The 1D averaging operator from cell-centers to nodes.
    """
    Av = sp.spdiags(
        (0.5 * np.ones((n, 1)) * [1, 1]).T, [-1, 0], n + 1, n, format="csr"
    ) + sp.csr_matrix(([0.5, 0.5], ([0, n], [0, n - 1])), shape=(n + 1, n))
    return Av


def ndgrid(*args, vector=True, order="F"):
    """Generate gridded locations for 1D, 2D, or 3D tensors.

    For 1D, 2D, or 3D tensors, this function takes the unique positions defining
    a tensor along each of its axis and returns the gridded locations.
    For 2D and 3D meshes, the user may treat the unique *x*, *y* (and *z*)
    positions a successive positional arguments or as a single argument using
    a list [*x*, *y*, (*z*)].

    For outputs, let *dim* be the number of dimension (1, 2 or 3) and let *n* be
    the total number of gridded locations. The gridded *x*, *y* (and *z*) locations
    can be return as a single numpy array of shape [n, ndim]. The user can also
    return the gridded *x*, *y* (and *z*) locations as a list of length *ndim*.
    The list contains entries contain the *x*, *y* (and *z*) locations as tensors.
    See examples.

    Parameters
    ----------
    *args : (n, dim) numpy.ndarray or (dim) list of (n) numpy.ndarray
        Positions along each axis of the tensor. The user can define these as
        successive positional arguments *x*, *y*, (and *z*) or as a single argument
        using a list [*x*, *y*, (*z*)].
    vector : bool, optional
        If *True*, the output is a numpy array of dimension [n, ndim]. If *False*,
        the gridded x, y (and z) locations are returned as separate ndarrays in a list.
        Default is *True*.
    order : {'F', 'C', 'A'}
        Define ordering using one of the following options:
        'C' is C-like ordering, 'F' is Fortran-like ordering, 'A' is Fortran
        ordering if memory is contigious and C-like otherwise. Default = 'F'.
        See :func:`numpy.reshape` for more on this argument.

    Returns
    -------
    numpy.ndarray or list of numpy.ndarray
        If *vector* = *True* the gridded *x*, *y*, (and *z*) locations are
        returned as a numpy array of shape [n, ndim]. If *vector* = *False*,
        the gridded *x*, *y*, (and *z*) are returned as a list of vectors.

    Examples
    --------
    >>> from discretize.utils import ndgrid
    >>> import numpy as np

    >>> x = np.array([1, 2, 3])
    >>> y = np.array([2, 4])

    >>> ndgrid([x, y])
    array([[1, 2],
           [2, 2],
           [3, 2],
           [1, 4],
           [2, 4],
           [3, 4]])

    >>> ndgrid(x, y, order='C')
    array([[1, 2],
           [1, 4],
           [2, 2],
           [2, 4],
           [3, 2],
           [3, 4]])

    >>> ndgrid(x, y, vector=False)
    [array([[1, 1],
           [2, 2],
           [3, 3]]), array([[2, 4],
           [2, 4],
           [2, 4]])]
    """
    # Read the keyword arguments, and only accept a vector=True/False
    if not isinstance(vector, bool):
        raise TypeError("'vector' keyword must be a bool")

    # you can either pass a list [x1, x2, x3] or each seperately
    if isinstance(args[0], list):
        xin = args[0]
    else:
        xin = args

    # Each vector needs to be a numpy array
    try:
        if len(xin) == 1:
            return np.array(xin[0])
        meshed = np.meshgrid(*xin, indexing="ij")
    except Exception:
        raise TypeError("All arguments must be array like")

    if vector:
        return np.column_stack([x.reshape(-1, order=order) for x in meshed])
    return meshed


def make_boundary_bool(shape, bdir="xyz", **kwargs):
    r"""Return boundary indices of a tensor grid.

    For a tensor grid whose shape is given (1D, 2D or 3D), this function
    returns a boolean index array identifying the x, y and/or z
    boundary locations.

    Parameters
    ----------
    shape : (dim) tuple of int
        Defines the shape of the tensor (1D, 2D or 3D).
    bdir : str containing characters 'x', 'y' and/or 'z'
        Specify the boundaries whose indices you want returned; e.g. for a 3D
        tensor, you may set *dir* = 'xz' to return the indices of the x and
        z boundary locations.

    Returns
    -------
    numpy.ndarray of bool
        Indices of boundary locations of the tensor for specified boundaries. The
        returned order matches the order the items occur in the flattened ``ndgrid``

    Examples
    --------
    Here we construct a 3x3 tensor and find the indices of the boundary locations.

    >>> from discretize.utils.matrix_utils import ndgrid, make_boundary_bool
    >>> import numpy as np

    Define a 3x3 tensor grid

    >>> x = np.array([1, 2, 3])
    >>> y = np.array([2, 4, 6])
    >>> tensor_grid = ndgrid(x, y)

    Find indices of boundary locations.

    >>> shape = (len(x), len(y))
    >>> bool_ind = make_boundary_bool(shape)
    >>> tensor_grid[bool_ind]
    array([[1, 2],
           [2, 2],
           [3, 2],
           [1, 4],
           [3, 4],
           [1, 6],
           [2, 6],
           [3, 6]])

    Find indices of locations of only the x boundaries,

    >>> bool_ind_x = make_boundary_bool(shape, 'x')
    >>> tensor_grid[bool_ind_x]
    array([[1, 2],
           [3, 2],
           [1, 4],
           [3, 4],
           [1, 6],
           [3, 6]])
    """
    old_dir = kwargs.pop("dir", None)
    if old_dir is not None:
        warnings.warn(
            "The `dir` keyword argument has been renamed to `bdir` to avoid shadowing the "
            "builtin variable `dir`. This will be removed in discretize 1.0.0",
            FutureWarning,
            stacklevel=2,
        )
        bdir = old_dir
    is_b = np.zeros(shape, dtype=bool, order="F")
    if "x" in bdir:
        is_b[[0, -1]] = True
    if len(shape) > 1:
        if "y" in bdir:
            is_b[:, [0, -1]] = True
    if len(shape) > 2:
        if "z" in bdir:
            is_b[:, :, [0, -1]] = True
    return is_b.reshape(-1, order="F")


def ind2sub(shape, inds):
    r"""Return subscripts of tensor grid elements from indices.

    This function is a wrapper for :func:`numpy.unravel_index` with a hard-coded Fortran
    order.

    Consider the :math:`n^{th}` element of a tensor grid with *N* elements.
    The position of this element in the tensor grid can also be defined by
    subscripts (i,j,k). For an array containing the indices for a set of
    tensor elements, this function returns the corresponding subscripts.

    Parameters
    ----------
    shape : (dim) tuple of int
        Defines the shape of the tensor (1D, 2D or 3D).
    inds : array_like of int
        The indices of the tensor elements whose subscripts you want returned.

    Returns
    -------
    (dim) tuple of numpy.ndarray
        Corresponding subscipts for the indices provided. The output is a
        tuple containing 1D integer arrays for the i, j and k subscripts, respectively.
        The output array will match the shape of the **inds** input.

    See Also
    --------
    numpy.unravel_index
    """
    return np.unravel_index(inds, shape, order="F")


def sub2ind(shape, subs):
    r"""Return indices of tensor grid elements from subscripts.

    This function is a wrapper for :func:`numpy.ravel_multi_index` with a hard-coded
    Fortran order, and a column order for the ``multi_index``

    Consider elements of a tensors grid whose positions are given by the
    subscripts (i,j,k). This function will return the corresponding indices
    of these elements. Each row of the input array *subs* defines the
    ijk for a particular tensor element.

    Parameters
    ----------
    shape : (dim) tuple of int
        Defines the shape of the tensor (1D, 2D or 3D).
    subs : (N, dim) array_like of int
        The subscripts of the tensor grid elements. Each rows defines the position
        of a particular tensor element. The shape of of the array is (N, ndim).

    Returns
    -------
    numpy.ndarray of int
        The indices of the tensor grid elements defined by *subs*.

    See Also
    --------
    numpy.ravel_multi_index

    Examples
    --------
    He we recreate the examples from :func:`numpy.ravel_multi_index` to illustrate the
    differences. The indices corresponding to each dimension are now columns in the
    array (instead of rows), and it assumed to use a Fortran order.

    >>> import numpy as np
    >>> from discretize.utils import sub2ind
    >>> arr = np.array([[3, 4], [6, 5], [6, 1]])
    >>> sub2ind((7, 6), arr)
    array([31, 41, 13], dtype=int64)
    """
    if len(shape) == 1:
        return subs
    subs = np.atleast_2d(subs)
    if subs.shape[1] != len(shape):
        raise ValueError(
            "Indexing must be done as a column vectors. e.g. [[3,6],[6,2],...]"
        )
    inds = np.ravel_multi_index(subs.T, shape, order="F")
    return mkvc(inds)


def get_subarray(A, ind):
    """Extract a subarray.

    For a :class:`numpy.ndarray`, the function **get_subarray** extracts a subset of
    the array. The portion of the original array being extracted is defined
    by providing the indices along each axis.

    Parameters
    ----------
    A : numpy.ndarray
        The original numpy array. Must be 1, 2 or 3 dimensions.
    ind : (dim) list of numpy.ndarray
        A list of numpy arrays containing the indices being extracted along each
        dimension. The length of the list must equal the dimensions of the input array.

    Returns
    -------
    numpy.ndarray
        The subarray extracted from the original array

    Examples
    --------
    Here we construct a random 3x3 numpy array and use **get_subarray** to extract
    the first column.

    >>> from discretize.utils import get_subarray
    >>> import numpy as np
    >>> rng = np.random.default_rng(421)
    >>> A = rng.random((3, 3))
    >>> A
    array([[1.07969034e-04, 9.78613931e-01, 6.62123429e-01],
           [8.80722877e-01, 7.61035691e-01, 7.42546796e-01],
           [9.09488911e-01, 7.80626334e-01, 8.67663825e-01]])

    Define the indexing along the columns and rows and create the indexing list

    >>> ind_x = np.array([0, 1, 2])
    >>> ind_y = np.array([0, 2])
    >>> ind = [ind_x, ind_y]

    Extract the first, and third column of A

    >>> get_subarray(A, ind)
    array([[1.07969034e-04, 6.62123429e-01],
           [8.80722877e-01, 7.42546796e-01],
           [9.09488911e-01, 8.67663825e-01]])
    """
    if not isinstance(ind, list):
        raise TypeError("ind must be a list of vectors")
    if len(A.shape) != len(ind):
        raise ValueError("ind must have the same length as the dimension of A")

    if len(A.shape) == 2:
        return A[ind[0], :][:, ind[1]]
    elif len(A.shape) == 3:
        return A[ind[0], :, :][:, ind[1], :][:, :, ind[2]]
    else:
        raise Exception("get_subarray does not support dimension asked.")


def inverse_3x3_block_diagonal(
    a11, a12, a13, a21, a22, a23, a31, a32, a33, return_matrix=True, **kwargs
):
    r"""Invert a set of 3x3 matricies from vectors containing their elements.

    Parameters
    ----------
    a11, a12, ..., a33 : (n_blocks) numpy.ndarray
        Vectors which contain the
        corresponding element for all 3x3 matricies
    return_matrix : bool, optional
        - **True**: Returns the sparse block 3x3 matrix *M* (default).
        - **False:** Returns the vectors containing the elements of each matrix' inverse.

    Returns
    -------
    (3 * n_blocks, 3 * n_blocks) scipy.sparse.coo_matrix or list of (n_blocks)
        numpy.ndarray. If *return_matrix = False*, the function will return vectors
        *b11, b12, b13, b21, b22, b23, b31, b32, b33*. If *return_matrix = True*, the
        function will return the block matrix *M*.

    Notes
    -----
    The elements of a 3x3 matrix *A* are given by:

    .. math::
        A = \begin{bmatrix}
        a_{11} & a_{12} & a_{13} \\
        a_{21} & a_{22} & a_{23} \\
        a_{31} & a_{32} & a_{33}
        \end{bmatrix}

    For a set of 3x3 matricies, the elements may be stored in a set of 9 distinct vectors
    :math:`\mathbf{a_{11}}`, :math:`\mathbf{a_{12}}`, ..., :math:`\mathbf{a_{33}}`.
    For each matrix, **inverse_3x3_block_diagonal** ouputs the vectors containing the
    elements of each matrix' inverse; i.e.
    :math:`\mathbf{b_{11}}`, :math:`\mathbf{b_{12}}`, ..., :math:`\mathbf{b_{33}}`
    where:

    .. math::
        A^{-1} = B = \begin{bmatrix}
        b_{11} & b_{12} & b_{13} \\
        b_{21} & b_{22} & b_{23} \\
        b_{31} & b_{32} & b_{33}
        \end{bmatrix}

    For special applications, we may want to output the elements of the inverses
    of the matricies as a 3x3 block matrix of the form:

    .. math::
        M = \begin{bmatrix}
        D_{11} & D_{12} & D_{13} \\
        D_{21} & D_{22} & D_{23} \\
        D_{31} & D_{32} & D_{33}
        \end{bmatrix}

    where :math:`D_{ij}` are diagonal matrices whose non-zero elements
    are defined by vector :math:`\\mathbf{b_{ij}}`. Where *n* is the
    number of matricies, the block matrix is sparse with dimensions
    (3n, 3n).

    Examples
    --------
    Here, we define four 3x3 matricies and reorganize their elements into
    9 vectors a11, a12, ..., a33. We then examine the outputs of the
    function **inverse_3x3_block_diagonal** when the argument
    *return_matrix* is set to both *True* and *False*.

    >>> from discretize.utils import inverse_3x3_block_diagonal
    >>> import numpy as np
    >>> import scipy as sp
    >>> import matplotlib.pyplot as plt

    Define four 3x3 matricies, and organize their elements into nine vectors

    >>> A1 = np.random.uniform(1, 10, (3, 3))
    >>> A2 = np.random.uniform(1, 10, (3, 3))
    >>> A3 = np.random.uniform(1, 10, (3, 3))
    >>> A4 = np.random.uniform(1, 10, (3, 3))
    >>> [[a11, a12, a13], [a21, a22, a23], [a31, a32, a33]] = np.stack(
    ...     [A1, A2, A3, A4], axis=-1
    ... )

    Return the elements of their inverse and validate

    >>> b11, b12, b13, b21, b22, b23, b31, b32, b33 = inverse_3x3_block_diagonal(
    ...     a11, a12, a13, a21, a22, a23, a31, a32, a33, return_matrix=False
    ... )
    >>> Bs = np.stack([[b11, b12, b13],[b21, b22, b23],[b31, b32, b33]])
    >>> B1, B2, B3, B4 = Bs.transpose((2, 0, 1))

    >>> np.linalg.inv(A1)
    array([[ 0.20941584,  0.18477151, -0.22637147],
           [-0.06420656, -0.34949639,  0.29216461],
           [-0.14226339,  0.11160555,  0.0907583 ]])
    >>> B1
    array([[ 0.20941584,  0.18477151, -0.22637147],
           [-0.06420656, -0.34949639,  0.29216461],
           [-0.14226339,  0.11160555,  0.0907583 ]])

    We can also return this as a sparse matrix with block diagonal inverse

    >>> M = inverse_3x3_block_diagonal(
    ...     a11, a12, a13, a21, a22, a23, a31, a32, a33
    ... )
    >>> plt.spy(M)
    >>> plt.show()
    """
    if "returnMatrix" in kwargs:
        warnings.warn(
            "The returnMatrix keyword argument has been deprecated, please use return_matrix. "
            "This will be removed in discretize 1.0.0",
            FutureWarning,
            stacklevel=2,
        )
        return_matrix = kwargs["returnMatrix"]

    a11 = mkvc(a11)
    a12 = mkvc(a12)
    a13 = mkvc(a13)
    a21 = mkvc(a21)
    a22 = mkvc(a22)
    a23 = mkvc(a23)
    a31 = mkvc(a31)
    a32 = mkvc(a32)
    a33 = mkvc(a33)

    detA = (
        a31 * a12 * a23
        - a31 * a13 * a22
        - a21 * a12 * a33
        + a21 * a13 * a32
        + a11 * a22 * a33
        - a11 * a23 * a32
    )

    b11 = +(a22 * a33 - a23 * a32) / detA
    b12 = -(a12 * a33 - a13 * a32) / detA
    b13 = +(a12 * a23 - a13 * a22) / detA

    b21 = +(a31 * a23 - a21 * a33) / detA
    b22 = -(a31 * a13 - a11 * a33) / detA
    b23 = +(a21 * a13 - a11 * a23) / detA

    b31 = -(a31 * a22 - a21 * a32) / detA
    b32 = +(a31 * a12 - a11 * a32) / detA
    b33 = -(a21 * a12 - a11 * a22) / detA

    if not return_matrix:
        return b11, b12, b13, b21, b22, b23, b31, b32, b33

    return sp.vstack(
        (
            sp.hstack((sdiag(b11), sdiag(b12), sdiag(b13))),
            sp.hstack((sdiag(b21), sdiag(b22), sdiag(b23))),
            sp.hstack((sdiag(b31), sdiag(b32), sdiag(b33))),
        )
    )


def inverse_2x2_block_diagonal(a11, a12, a21, a22, return_matrix=True, **kwargs):
    r"""
    Invert a set of 2x2 matricies from vectors containing their elements.

    Parameters
    ----------
    a11, a12, a21, a22 : (n_blocks) numpy.ndarray
        All arguments a11, a12, a21, a22 are vectors which contain the
        corresponding element for all 2x2 matricies
    return_matrix : bool, optional
        - **True:** Returns the sparse block 2x2 matrix *M*.
        - **False:** Returns the vectors containing the elements of each matrix' inverse.

    Returns
    -------
    (2 * n_blocks, 2 * n_blocks) scipy.sparse.coo_matrix or list of (n_blocks) numpy.ndarray
        If *return_matrix = False*, the function will return vectors
        *b11, b12, b21, b22*.
        If *return_matrix = True*, the function will return the
        block matrix *M*

    Notes
    -----
    The elements of a 2x2 matrix *A* are given by:

    .. math::
        A = \begin{bmatrix}
        a_{11} & a_{12} \\
        a_{21} & a_{22}
        \end{bmatrix}

    For a set of 2x2 matricies, the elements may be stored in a set of 4 distinct vectors
    :math:`\mathbf{a_{11}}`, :math:`\mathbf{a_{12}}`, :math:`\mathbf{a_{21}}` and
    :math:`\mathbf{a_{22}}`.
    For each matrix, **inverse_2x2_block_diagonal** ouputs the vectors containing the
    elements of each matrix' inverse; i.e.
    :math:`\mathbf{b_{11}}`, :math:`\mathbf{b_{12}}`, :math:`\mathbf{b_{21}}` and
    :math:`\mathbf{b_{22}}` where:

    .. math::
        A^{-1} = B = \begin{bmatrix}
        b_{11} & b_{12} \\
        b_{21} & b_{22}
        \end{bmatrix}

    For special applications, we may want to output the elements of the inverses
    of the matricies as a 2x2 block matrix of the form:

    .. math::
        M = \begin{bmatrix}
        D_{11} & D_{12} \\
        D_{21} & D_{22}
        \end{bmatrix}

    where :math:`D_{ij}` are diagonal matrices whose non-zero elements
    are defined by vector :math:`\mathbf{b_{ij}}`. Where *n* is the
    number of matricies, the block matrix is sparse with dimensions
    (2n, 2n).

    Examples
    --------
    Here, we define four 2x2 matricies and reorganize their elements into
    4 vectors a11, a12, a21 and a22. We then examine the outputs of the
    function **inverse_2x2_block_diagonal** when the argument
    *return_matrix* is set to both *True* and *False*.

    >>> from discretize.utils import inverse_2x2_block_diagonal
    >>> import numpy as np
    >>> import matplotlib.pyplot as plt

    Define four 3x3 matricies, and organize their elements into four vectors

    >>> A1 = np.random.uniform(1, 10, (2, 2))
    >>> A2 = np.random.uniform(1, 10, (2, 2))
    >>> A3 = np.random.uniform(1, 10, (2, 2))
    >>> A4 = np.random.uniform(1, 10, (2, 2))
    >>> [[a11, a12], [a21, a22]] = np.stack([A1, A2, A3, A4], axis=-1)

    Return the elements of their inverse and validate

    >>> b11, b12, b21, b22 = inverse_2x2_block_diagonal(
    ...     a11, a12, a21, a22, return_matrix=False
    ... )
    >>> Bs = np.stack([[b11, b12],[b21, b22]])
    >>> B1, B2, B3, B4 = Bs.transpose((2, 0, 1))

    >>> np.linalg.inv(A1)
    array([[ 0.34507439, -0.4831833 ],
           [-0.24286626,  0.57531461]])
    >>> B1
    array([[ 0.34507439, -0.4831833 ],
           [-0.24286626,  0.57531461]])

    Plot the sparse block matrix containing elements of the inverses

    >>> M = inverse_2x2_block_diagonal(
    ...     a11, a12, a21, a22
    ... )
    >>> plt.spy(M)
    >>> plt.show()
    """
    if "returnMatrix" in kwargs:
        warnings.warn(
            "The returnMatrix keyword argument has been deprecated, please use return_matrix. "
            "This will be removed in discretize 1.0.0",
            FutureWarning,
            stacklevel=2,
        )
        return_matrix = kwargs["returnMatrix"]

    a11 = mkvc(a11)
    a12 = mkvc(a12)
    a21 = mkvc(a21)
    a22 = mkvc(a22)

    # compute inverse of the determinant.
    detAinv = 1.0 / (a11 * a22 - a21 * a12)

    b11 = +detAinv * a22
    b12 = -detAinv * a12
    b21 = -detAinv * a21
    b22 = +detAinv * a11

    if not return_matrix:
        return b11, b12, b21, b22

    return sp.vstack(
        (sp.hstack((sdiag(b11), sdiag(b12))), sp.hstack((sdiag(b21), sdiag(b22))))
    )


def invert_blocks(A):
    """Invert a set of 2x2 or 3x3 matricies.

    This is a shortcut function that will only invert 2x2 and 3x3 matrices.
    The function is broadcast over the last two dimensions of A.

    Parameters
    ----------
    A : (..., N, N) numpy.ndarray
        the block of matrices to invert, N must be either 2 or 3.

    Returns
    -------
    (..., N, N) numpy.ndarray
        the block of inverted matrices

    See Also
    --------
    numpy.linalg.inv : Similar to this function, but is not specialized to 2x2 or 3x3
    inverse_2x2_block_diagonal : use when each element of the blocks is separated
    inverse_3x3_block_diagonal : use when each element of the blocks is separated

    Examples
    --------
    >>> from discretize.utils import invert_blocks
    >>> import numpy as np
    >>> x = np.ones((1000, 3, 3))
    >>> x[..., 1, 1] = 0
    >>> x[..., 1, 2] = 0
    >>> x[..., 2, 1] = 0
    >>> As = np.einsum('...ij,...jk', x, x.transpose(0, 2, 1))
    >>> Ainvs = invert_blocks(As)
    >>> As[0] @ Ainvs[0]
    array([[1., 0., 0.],
           [0., 1., 0.],
           [0., 0., 1.]])
    """
    if A.shape[-1] != A.shape[-2]:
        raise ValueError(f"Last two dimensions are not equal, got {A.shape}")

    if A.shape[-1] == 2:
        a11 = A[..., 0, 0]
        a12 = A[..., 0, 1]
        a21 = A[..., 1, 0]
        a22 = A[..., 1, 1]

        detA = a11 * a22 - a21 * a12
        B = np.empty_like(A)
        B[..., 0, 0] = a22 / detA
        B[..., 0, 1] = -a12 / detA
        B[..., 1, 0] = -a21 / detA
        B[..., 1, 1] = a11 / detA

    elif A.shape[-1] == 3:
        a11 = A[..., 0, 0]
        a12 = A[..., 0, 1]
        a13 = A[..., 0, 2]
        a21 = A[..., 1, 0]
        a22 = A[..., 1, 1]
        a23 = A[..., 1, 2]
        a31 = A[..., 2, 0]
        a32 = A[..., 2, 1]
        a33 = A[..., 2, 2]

        B = np.empty_like(A)
        B[..., 0, 0] = a22 * a33 - a23 * a32
        B[..., 0, 1] = a13 * a32 - a12 * a33
        B[..., 0, 2] = a12 * a23 - a13 * a22

        B[..., 1, 0] = a31 * a23 - a21 * a33
        B[..., 1, 1] = a11 * a33 - a31 * a13
        B[..., 1, 2] = a21 * a13 - a11 * a23

        B[..., 2, 0] = a21 * a32 - a31 * a22
        B[..., 2, 1] = a31 * a12 - a11 * a32
        B[..., 2, 2] = a11 * a22 - a21 * a12

        detA = a11 * B[..., 0, 0] + a21 * B[..., 0, 1] + a31 * B[..., 0, 2]
        B /= detA[..., None, None]
    else:
        raise NotImplementedError("Only supports 2x2 and 3x3 blocks")
    return B


class TensorType(object):
    r"""Class for determining property tensor type.

    For a given *mesh*, the **TensorType** class examines the :class:`numpy.ndarray`
    *tensor* to determine whether *tensor* defines a scalar, isotropic,
    diagonal anisotropic or full tensor anisotropic constitutive relationship
    for each cell on the mesh. The general theory behind this functionality
    is explained below.

    Parameters
    ----------
    mesh : discretize.base.BaseTensorMesh
        An instance of any of the mesh classes support in discretize; i.e. *TensorMesh*,
        *CylindricalMesh*, *TreeMesh* or *CurvilinearMesh*.
    tensor : numpy.ndarray or a float
        The shape of the input argument *tensor* must fall into one of these
        classifications:

        - *Scalar:* A float is entered.
        - *Isotropic:* A 1D numpy.ndarray with a property value for every cell.
        - *Anisotropic:* A (*nCell*, *dim*) numpy.ndarray of shape where each row
          defines the diagonal-anisotropic property parameters for each cell.
          *nParam* = 2 for 2D meshes and *nParam* = 3 for 3D meshes.
        - *Tensor:* A (*nCell*, *nParam*) numpy.ndarray where each row
          defines the full anisotropic property parameters for each cell.
          *nParam* = 3 for 2D meshes and *nParam* = 6 for 3D meshes.

    Notes
    -----
    The relationship between a quantity and its response to external
    stimuli (e.g. Ohm's law) can be defined by a scalar quantity:

    .. math::
        \vec{j} = \sigma \vec{e}

    Or in the case of anisotropy, the relationship is defined generally by
    a symmetric tensor:

    .. math::
        \vec{j} = \Sigma \vec{e} \;\;\; where \;\;\;
        \Sigma = \begin{bmatrix}
        \sigma_{xx} & \sigma_{xy} & \sigma_{xz} \\
        \sigma_{xy} & \sigma_{yy} & \sigma_{yz} \\
        \sigma_{xz} & \sigma_{yz} & \sigma_{zz}
        \end{bmatrix}

    In 3D, the tensor is defined by 6 independent element (3 independent elements in
    2D). When using the input argument *tensor* to define the consitutive relationship
    for every cell in the *mesh*, there are 4 classifications recognized by discretize:

    - **Scalar:** :math:`\vec{j} = \sigma \vec{e}`, where :math:`\sigma` a constant.
      Thus the input argument *tensor* is a float.
    - **Isotropic:** :math:`\vec{j} = \sigma \vec{e}`, where :math:`\sigma` varies
      spatially. Thus the input argument *tensor* is a 1D array that provides a
      :math:`\sigma` value for every cell in the mesh.
    - **Anisotropic:** :math:`\vec{j} = \Sigma \vec{e}`, where the off-diagonal elements
      are zero. That is, :math:`\Sigma` is diagonal. In this case, the input argument
      *tensor* defining the physical properties in each cell is a :class:`numpy.ndarray`
      of shape (*nCells*, *dim*).
    - **Tensor:** :math:`\vec{j} = \Sigma \vec{e}`, where off-diagonal elements are
      non-zero and :math:`\Sigma` is a full tensor. In this case, the input argument
      *tensor* defining the physical properties in each cell is a :class:`numpy.ndarray`
      of shape (*nCells*, *nParam*). In 2D, *nParam* = 3 and in 3D, *nParam* = 6.
    """

    def __init__(self, mesh, tensor):
        if tensor is None:  # default is ones
            self._tt = -1
            self._tts = "none"
        elif is_scalar(tensor):
            self._tt = 0
            self._tts = "scalar"
        elif tensor.size == mesh.nC:
            self._tt = 1
            self._tts = "isotropic"
        elif (mesh.dim == 2 and tensor.size == mesh.nC * 2) or (
            mesh.dim == 3 and tensor.size == mesh.nC * 3
        ):
            self._tt = 2
            self._tts = "anisotropic"
        elif (mesh.dim == 2 and tensor.size == mesh.nC * 3) or (
            mesh.dim == 3 and tensor.size == mesh.nC * 6
        ):
            self._tt = 3
            self._tts = "tensor"
        else:
            raise Exception("Unexpected shape of tensor: {}".format(tensor.shape))

    def __str__(self):
        """Represent tensor type as a string."""
        return "TensorType[{0:d}]: {1!s}".format(self._tt, self._tts)

    def __eq__(self, v):
        """Compare tensor type equal to a value."""
        return self._tt == v

    def __le__(self, v):
        """Compare tensor type less than or equal to a value."""
        return self._tt <= v

    def __ge__(self, v):
        """Compare tensor type greater than or equal to a value."""
        return self._tt >= v

    def __lt__(self, v):
        """Compare tensor type less than a value."""
        return self._tt < v

    def __gt__(self, v):
        """Compare tensor type greater than a value."""
        return self._tt > v


def make_property_tensor(mesh, tensor):
    r"""Construct the physical property tensor.

    For a given *mesh*, the input parameter *tensor* is a :class:`numpy.ndarray`
    defining the constitutive relationship (e.g. Ohm's law) between two
    discrete vector quantities :math:`\boldsymbol{j}` and
    :math:`\boldsymbol{e}` living at cell centers. The function
    **make_property_tensor** constructs the property tensor
    :math:`\boldsymbol{M}` for the entire mesh such that:

    >>> j = M @ e

    where the Cartesian components of the discrete vector for
    are organized according to:

    >>> e = np.r_[ex, ey, ez]
    >>> j = np.r_[jx, jy, jz]

    Parameters
    ----------
    mesh : discretize.base.BaseMesh
       A mesh
    tensor : numpy.ndarray or a float
        - *Scalar:* A float is entered.
        - *Isotropic:* A 1D numpy.ndarray with a property value for every cell.
        - *Anisotropic:* A (*nCell*, *dim*) numpy.ndarray where each row
          defines the diagonal-anisotropic property parameters for each cell.
          *nParam* = 2 for 2D meshes and *nParam* = 3 for 3D meshes.
        - *Tensor:* A (*nCell*, *nParam*) numpy.ndarray where each row defines
          the full anisotropic property parameters for each cell. *nParam* = 3 for 2D
          meshes and *nParam* = 6 for 3D meshes.

    Returns
    -------
    (dim * n_cells, dim * n_cells) scipy.sparse.coo_matrix
        The property tensor.

    Notes
    -----
    The relationship between a quantity and its response to external
    stimuli (e.g. Ohm's law) in each cell can be defined by a scalar
    function :math:`\sigma` in the isotropic case, or by a tensor
    :math:`\Sigma` in the anisotropic case, i.e.:

    .. math::
        \vec{j} = \sigma \vec{e} \;\;\;\;\;\; \textrm{or} \;\;\;\;\;\; \vec{j} = \Sigma \vec{e}

    where

    .. math::
        \Sigma = \begin{bmatrix}
        \sigma_{xx} & \sigma_{xy} & \sigma_{xz} \\
        \sigma_{xy} & \sigma_{yy} & \sigma_{yz} \\
        \sigma_{xz} & \sigma_{yz} & \sigma_{zz}
        \end{bmatrix}

    Examples
    --------
    For the 4 classifications allowable (scalar, isotropic, anistropic and tensor),
    we construct and compare the property tensor on a small 2D mesh. For this
    example, note the following:

        - The dimensions for all property tensors are the same
        - All property tensors, except in the case of full anisotropy are diagonal
          sparse matrices
        - For the scalar case, the non-zero elements are equal
        - For the isotropic case, the non-zero elements repreat in order for the x, y
          (and z) components
        - For the anisotropic case (diagonal anisotropy), the non-zero elements do not
          repeat
        - For the tensor caes (full anisotropy), there are off-diagonal components

    >>> from discretize.utils import make_property_tensor
    >>> from discretize import TensorMesh
    >>> import numpy as np
    >>> import matplotlib.pyplot as plt
    >>> import matplotlib as mpl
    >>> rng = np.random.default_rng(421)

    Define a 2D tensor mesh

    >>> h = [1., 1., 1.]
    >>> mesh = TensorMesh([h, h], origin='00')

    Define a physical property for all cases (2D)

    >>> sigma_scalar = 4.
    >>> sigma_isotropic = rng.integers(1, 10, mesh.nC)
    >>> sigma_anisotropic = rng.integers(1, 10, (mesh.nC, 2))
    >>> sigma_tensor = rng.integers(1, 10, (mesh.nC, 3))

    Construct the property tensor in each case

    >>> M_scalar = make_property_tensor(mesh, sigma_scalar)
    >>> M_isotropic = make_property_tensor(mesh, sigma_isotropic)
    >>> M_anisotropic = make_property_tensor(mesh, sigma_anisotropic)
    >>> M_tensor = make_property_tensor(mesh, sigma_tensor)

    Plot the property tensors.

    >>> M_list = [M_scalar, M_isotropic, M_anisotropic, M_tensor]
    >>> case_list = ['Scalar', 'Isotropic', 'Anisotropic', 'Full Tensor']
    >>> ax1 = 4*[None]
    >>> fig = plt.figure(figsize=(15, 4))
    >>> for ii in range(0, 4):
    ...     ax1[ii] = fig.add_axes([0.05+0.22*ii, 0.05, 0.18, 0.9])
    ...     ax1[ii].imshow(
    ...         M_list[ii].todense(), interpolation='none', cmap='binary', vmax=10.
    ...     )
    ...     ax1[ii].set_title(case_list[ii], fontsize=24)
    >>> ax2 = fig.add_axes([0.92, 0.15, 0.01, 0.7])
    >>> norm = mpl.colors.Normalize(vmin=0., vmax=10.)
    >>> cbar = mpl.colorbar.ColorbarBase(
    ...     ax2, norm=norm, orientation="vertical", cmap=mpl.cm.binary
    ... )
    >>> plt.show()
    """
    if tensor is None:  # default is ones
        tensor = np.ones(mesh.nC)

    if is_scalar(tensor):
        tensor = tensor * np.ones(mesh.nC)

    propType = TensorType(mesh, tensor)
    if propType == 1:  # Isotropic!
        Sigma = sp.kron(sp.identity(mesh.dim), sdiag(mkvc(tensor)))
    elif propType == 2:  # Diagonal tensor
        Sigma = sdiag(mkvc(tensor))
    elif mesh.dim == 2 and tensor.size == mesh.nC * 3:  # Fully anisotropic, 2D
        tensor = tensor.reshape((mesh.nC, 3), order="F")
        row1 = sp.hstack((sdiag(tensor[:, 0]), sdiag(tensor[:, 2])))
        row2 = sp.hstack((sdiag(tensor[:, 2]), sdiag(tensor[:, 1])))
        Sigma = sp.vstack((row1, row2))
    elif mesh.dim == 3 and tensor.size == mesh.nC * 6:  # Fully anisotropic, 3D
        tensor = tensor.reshape((mesh.nC, 6), order="F")
        row1 = sp.hstack(
            (sdiag(tensor[:, 0]), sdiag(tensor[:, 3]), sdiag(tensor[:, 4]))
        )
        row2 = sp.hstack(
            (sdiag(tensor[:, 3]), sdiag(tensor[:, 1]), sdiag(tensor[:, 5]))
        )
        row3 = sp.hstack(
            (sdiag(tensor[:, 4]), sdiag(tensor[:, 5]), sdiag(tensor[:, 2]))
        )
        Sigma = sp.vstack((row1, row2, row3))
    else:
        raise Exception("Unexpected shape of tensor")

    return Sigma


def inverse_property_tensor(mesh, tensor, return_matrix=False, **kwargs):
    r"""Construct the inverse of the physical property tensor.

    For a given *mesh*, the input parameter *tensor* is a :class:`numpy.ndarray`
    defining the constitutive relationship (e.g. Ohm's law) between two
    discrete vector quantities :math:`\boldsymbol{j}` and
    :math:`\boldsymbol{e}` living at cell centers. Where :math:`\boldsymbol{M}`
    is the physical property tensor, **inverse_property_tensor**
    explicitly constructs the inverse of the physical
    property tensor :math:`\boldsymbol{M^{-1}}` for all cells such that:

    >>> e = Mi @ j

    where the Cartesian components of the discrete vectors are
    organized according to:

    >>> j = np.r_[jx, jy, jz]
    >>> e = np.r_[ex, ey, ez]

    Parameters
    ----------
    mesh : discretize.base.BaseMesh
       A mesh
    tensor : numpy.ndarray or float
        - *Scalar:* A float is entered.
        - *Isotropic:* A 1D numpy.ndarray with a property value for every cell.
        - *Anisotropic:* A (*nCell*, *dim*) numpy.ndarray where each row
          defines the diagonal-anisotropic property parameters for each cell.
          *nParam* = 2 for 2D meshes and *nParam* = 3 for 3D meshes.
        - *Tensor:* A (*nCell*, *nParam*) numpy.ndarray where each row defines
          the full anisotropic property parameters for each cell. *nParam* = 3 for 2D
          meshes and *nParam* = 6 for 3D meshes.

    return_matrix : bool, optional
        - *True:* the function returns the inverse of the property tensor.
        - *False:* the function returns the non-zero elements of the inverse of the
          property tensor in a numpy.ndarray in the same order as the input argument
          *tensor*.

    Returns
    -------
    numpy.ndarray or scipy.sparse.coo_matrix
        - If *return_matrix* = *False*, the function outputs the parameters defining the
          inverse of the property tensor in a numpy.ndarray with the same dimensions as
          the input argument *tensor*
        - If *return_natrix* = *True*, the function outputs the inverse of the property
          tensor as a *scipy.sparse.coo_matrix*.

    Notes
    -----
    The relationship between a quantity and its response to external
    stimuli (e.g. Ohm's law) in each cell can be defined by a scalar
    function :math:`\sigma` in the isotropic case, or by a tensor
    :math:`\Sigma` in the anisotropic case, i.e.:

    .. math::
        \vec{j} = \sigma \vec{e} \;\;\;\;\;\; \textrm{or} \;\;\;\;\;\;
        \vec{j} = \Sigma \vec{e}

    where

    .. math::
        \Sigma = \begin{bmatrix}
        \sigma_{xx} & \sigma_{xy} & \sigma_{xz} \\
        \sigma_{xy} & \sigma_{yy} & \sigma_{yz} \\
        \sigma_{xz} & \sigma_{yz} & \sigma_{zz}
        \end{bmatrix}

    Examples
    --------
    For the 4 classifications allowable (scalar, isotropic, anistropic and tensor),
    we construct the property tensor on a small 2D mesh. We then construct the
    inverse of the property tensor and compare.

    >>> from discretize.utils import make_property_tensor, inverse_property_tensor
    >>> from discretize import TensorMesh
    >>> import numpy as np
    >>> import matplotlib.pyplot as plt
    >>> import matplotlib as mpl
    >>> rng = np.random.default_rng(421)

    Define a 2D tensor mesh

    >>> h = [1., 1., 1.]
    >>> mesh = TensorMesh([h, h], origin='00')

    Define a physical property for all cases (2D)

    >>> sigma_scalar = 4.
    >>> sigma_isotropic = rng.integers(1, 10, mesh.nC)
    >>> sigma_anisotropic = rng.integers(1, 10, (mesh.nC, 2))
    >>> sigma_tensor = rng.integers(1, 10, (mesh.nC, 3))

    Construct the property tensor in each case

    >>> M_scalar = make_property_tensor(mesh, sigma_scalar)
    >>> M_isotropic = make_property_tensor(mesh, sigma_isotropic)
    >>> M_anisotropic = make_property_tensor(mesh, sigma_anisotropic)
    >>> M_tensor = make_property_tensor(mesh, sigma_tensor)

    Construct the inverse property tensor in each case

    >>> Minv_scalar = inverse_property_tensor(mesh, sigma_scalar, return_matrix=True)
    >>> Minv_isotropic = inverse_property_tensor(mesh, sigma_isotropic, return_matrix=True)
    >>> Minv_anisotropic = inverse_property_tensor(mesh, sigma_anisotropic, return_matrix=True)
    >>> Minv_tensor = inverse_property_tensor(mesh, sigma_tensor, return_matrix=True)

    Plot the property tensors.

    >>> M_list = [M_scalar, M_isotropic, M_anisotropic, M_tensor]
    >>> Minv_list = [Minv_scalar, Minv_isotropic, Minv_anisotropic, Minv_tensor]
    >>> case_list = ['Scalar', 'Isotropic', 'Anisotropic', 'Full Tensor']
    >>> fig1 = plt.figure(figsize=(15, 4))
    >>> ax1 = 4*[None]
    >>> for ii in range(0, 4):
    ...     ax1[ii] = fig1.add_axes([0.05+0.22*ii, 0.05, 0.18, 0.9])
    ...     ax1[ii].imshow(
    ...         M_list[ii].todense(), interpolation='none', cmap='binary', vmax=10.
    ...     )
    ...     ax1[ii].set_title('$M$ (' + case_list[ii] + ')', fontsize=24)
    >>> cax1 = fig1.add_axes([0.92, 0.15, 0.01, 0.7])
    >>> norm1 = mpl.colors.Normalize(vmin=0., vmax=10.)
    >>> cbar1 = mpl.colorbar.ColorbarBase(
    ...     cax1, norm=norm1, orientation="vertical", cmap=mpl.cm.binary
    ... )
    >>> plt.show()

    Plot the inverse property tensors.

    >>> fig2 = plt.figure(figsize=(15, 4))
    >>> ax2 = 4*[None]
    >>> for ii in range(0, 4):
    ...     ax2[ii] = fig2.add_axes([0.05+0.22*ii, 0.05, 0.18, 0.9])
    ...     ax2[ii].imshow(
    ...         Minv_list[ii].todense(), interpolation='none', cmap='binary', vmax=1.
    ...     )
    ...     ax2[ii].set_title('$M^{-1}$ (' + case_list[ii] + ')', fontsize=24)
    >>> cax2 = fig2.add_axes([0.92, 0.15, 0.01, 0.7])
    >>> norm2 = mpl.colors.Normalize(vmin=0., vmax=1.)
    >>> cbar2 = mpl.colorbar.ColorbarBase(
    ...     cax2, norm=norm2, orientation="vertical", cmap=mpl.cm.binary
    ... )
    >>> plt.show()
    """
    if "returnMatrix" in kwargs:
        warnings.warn(
            "The returnMatrix keyword argument has been deprecated, please use return_matrix. "
            "This will be removed in discretize 1.0.0",
            FutureWarning,
            stacklevel=2,
        )
        return_matrix = kwargs["returnMatrix"]

    propType = TensorType(mesh, tensor)

    if is_scalar(tensor):
        T = 1.0 / tensor
    elif propType < 3:  # Isotropic or Diagonal
        T = 1.0 / mkvc(tensor)  # ensure it is a vector.
    elif mesh.dim == 2 and tensor.size == mesh.nC * 3:  # Fully anisotropic, 2D
        tensor = tensor.reshape((mesh.nC, 3), order="F")
        B = inverse_2x2_block_diagonal(
            tensor[:, 0], tensor[:, 2], tensor[:, 2], tensor[:, 1], return_matrix=False
        )
        b11, b12, b21, b22 = B
        T = np.r_[b11, b22, b12]
    elif mesh.dim == 3 and tensor.size == mesh.nC * 6:  # Fully anisotropic, 3D
        tensor = tensor.reshape((mesh.nC, 6), order="F")
        B = inverse_3x3_block_diagonal(
            tensor[:, 0],
            tensor[:, 3],
            tensor[:, 4],
            tensor[:, 3],
            tensor[:, 1],
            tensor[:, 5],
            tensor[:, 4],
            tensor[:, 5],
            tensor[:, 2],
            return_matrix=False,
        )
        b11, b12, b13, b21, b22, b23, b31, b32, b33 = B
        T = np.r_[b11, b22, b33, b12, b13, b23]
    else:
        raise Exception("Unexpected shape of tensor")

    if return_matrix:
        return make_property_tensor(mesh, T)

    return T


def cross2d(x, y):
    """Compute the cross product of two vectors.

    This function will calculate the cross product as if
    the third component of each of these vectors was zero.

    The returned direction is perpendicular to both inputs,
    making it be solely in the third dimension.

    Parameters
    ----------
    x, y : array_like
        The vectors for the cross product.

    Returns
    -------
    x_cross_y : numpy.ndarray
        The cross product of x and y.
    """
    x = np.asarray(x)
    y = np.asarray(y)
    # np.cross(x, y) is deprecated for 2D input
    return x[..., 0] * y[..., 1] - x[..., 1] * y[..., 0]


class Zero(object):
    """Carries out arithmetic operations between 0 and arbitrary quantities.

    This class was designed to manage basic arithmetic operations between
    0 and :class:`numpy.ndarray` of any shape. It is a short circuiting evaluation that
    will return the expected values.

    Examples
    --------
    >>> import numpy as np
    >>> from discretize.utils import Zero
    >>> Z = Zero()
    >>> Z
    Zero
    >>> x = np.arange(5)
    >>> x + Z
    array([0, 1, 2, 3, 4])
    >>> Z - x
    array([ 0, -1, -2, -3, -4])
    >>> Z * x
    Zero
    >>> Z @ x
    Zero
    >>> Z[0]
    Zero
    """

    __numpy_ufunc__ = True
    __array_ufunc__ = None

    def __repr__(self):
        """Represent zeros a string."""
        return "Zero"

    def __add__(self, v):
        """Add a value to zero."""
        return v

    def __radd__(self, v):
        """Add zero to a value."""
        return v

    def __iadd__(self, v):
        """Add zero to a value inplace."""
        return v

    def __sub__(self, v):
        """Subtract a value from zero."""
        return -v

    def __rsub__(self, v):
        """Subtract zero from a value."""
        return v

    def __isub__(self, v):
        """Subtract zero from a value inplace."""
        return v

    def __mul__(self, v):
        """Multiply zero by a value."""
        return self

    def __rmul__(self, v):
        """Multiply a value by zero."""
        return self

    def __matmul__(self, v):
        """Multiply zero by a matrix."""
        return self

    def __rmatmul__(self, v):
        """Multiply a matrix by zero."""
        return self

    def __div__(self, v):
        """Divide zero by a value."""
        return self

    def __truediv__(self, v):
        """Divide zero by a value."""
        return self

    def __rdiv__(self, v):
        """Try to divide a value by zero."""
        raise ZeroDivisionError("Cannot divide by zero.")

    def __rtruediv__(self, v):
        """Try to divide a value by zero."""
        raise ZeroDivisionError("Cannot divide by zero.")

    def __rfloordiv__(self, v):
        """Try to divide a value by zero."""
        raise ZeroDivisionError("Cannot divide by zero.")

    def __pos__(self):
        """Return zero."""
        return self

    def __neg__(self):
        """Negate zero."""
        return self

    def __lt__(self, v):
        """Compare less than zero."""
        return 0 < v

    def __le__(self, v):
        """Compare less than or equal to zero."""
        return 0 <= v

    def __eq__(self, v):
        """Compare equal to zero."""
        return v == 0

    def __ne__(self, v):
        """Compare not equal to zero."""
        return not (0 == v)

    def __ge__(self, v):
        """Compare greater than or equal to zero."""
        return 0 >= v

    def __gt__(self, v):
        """Compare greater than zero."""
        return 0 > v

    def transpose(self):
        """Return the transpose of the *Zero* class, i.e. itself."""
        return self

    def __getitem__(self, key):
        """Get an element of the *Zero* class, i.e. itself."""
        return self

    @property
    def ndim(self):
        """Return the dimension of *Zero* class, i.e. *None*."""
        return None

    @property
    def shape(self):
        """Return the shape *Zero* class, i.e. *None*."""
        return _inftup(None)

    @property
    def T(self):
        """Return the *Zero* class as an operator."""
        return self


class Identity(object):
    """Carries out arithmetic operations involving the identity.

    This class was designed to manage basic arithmetic operations between the identity
    matrix and :class:`numpy.ndarray` of any shape. It is a short circuiting evaluation
    that will return the expected values.

    Parameters
    ----------
    positive : bool, optional
        Whether it is a positive (or negative) Identity matrix

    Examples
    --------
    >>> import numpy as np
    >>> from discretize.utils import Identity, Zero
    >>> Z = Zero()
    >>> I = Identity()
    >>> x = np.arange(5)
    >>> x + I
    array([1, 2, 3, 4, 5])
    >>> I - x
    array([ 1, 0, -1, -2, -3])
    >>> I * x
    array([0, 1, 2, 3, 4])
    >>> I @ x
    array([0, 1, 2, 3, 4])
    >>> I @ Z
    Zero
    """

    __numpy_ufunc__ = True
    __array_ufunc__ = None

    _positive = True

    def __init__(self, positive=True):
        self._positive = positive

    def __repr__(self):
        """Represent 1 (or -1 if not positive)."""
        if self._positive:
            return "I"
        else:
            return "-I"

    def __pos__(self):
        """Return positive 1 (or -1 if not positive)."""
        return self

    def __neg__(self):
        """Negate 1 (or -1 if not positive)."""
        return Identity(not self._positive)

    def __add__(self, v):
        """Add 1 (or -1 if not positive) to a value."""
        if sp.issparse(v):
            return v + speye(v.shape[0]) if self._positive else v - speye(v.shape[0])
        return v + 1 if self._positive else v - 1

    def __radd__(self, v):
        """Add 1 (or -1 if not positive) to a value."""
        return self.__add__(v)

    def __sub__(self, v):
        """Subtract a value from 1 (or -1 if not positive)."""
        return self + -v

    def __rsub__(self, v):
        """Subtract 1 (or -1 if not positive) from a value."""
        return -self + v

    def __mul__(self, v):
        """Multiply 1 (or -1 if not positive) by a value."""
        return v if self._positive else -v

    def __rmul__(self, v):
        """Multiply 1 (or -1 if not positive) by a value."""
        return v if self._positive else -v

    def __matmul__(self, v):
        """Multiply 1 (or -1 if not positive) by a matrix."""
        return v if self._positive else -v

    def __rmatmul__(self, v):
        """Multiply a matrix by 1 (or -1 if not positive)."""
        return v if self._positive else -v

    def __div__(self, v):
        """Divide 1 (or -1 if not positive) by a value."""
        if sp.issparse(v):
            raise NotImplementedError("Sparse arrays not divisibile.")
        return 1 / v if self._positive else -1 / v

    def __truediv__(self, v):
        """Divide 1 (or -1 if not positive) by a value."""
        if sp.issparse(v):
            raise NotImplementedError("Sparse arrays not divisibile.")
        return 1.0 / v if self._positive else -1.0 / v

    def __rdiv__(self, v):
        """Divide a value by 1 (or -1 if not positive)."""
        return v if self._positive else -v

    def __rtruediv__(self, v):
        """Divide a value by 1 (or -1 if not positive)."""
        return v if self._positive else -v

    def __floordiv__(self, v):
        """Flooring division of 1 (or -1 if not positive) by a value."""
        return 1 // v if self._positive else -1 // v

    def __rfloordiv__(self, v):
        """Flooring division of a value by 1 (or -1 if not positive)."""
        return v // 1 if self._positive else v // -1

    def __lt__(self, v):
        """Compare less than 1 (or -1 if not positive)."""
        return 1 < v if self._positive else -1 < v

    def __le__(self, v):
        """Compare less than or equal to 1 (or -1 if not positive)."""
        return 1 <= v if self._positive else -1 <= v

    def __eq__(self, v):
        """Compare equal to 1 (or -1 if not positive)."""
        return v == 1 if self._positive else v == -1

    def __ne__(self, v):
        """Compare not equal to 1 (or -1 if not positive)."""
        return (not (1 == v)) if self._positive else (not (-1 == v))

    def __ge__(self, v):
        """Compare greater than or equal to 1 (or -1 if not positive)."""
        return 1 >= v if self._positive else -1 >= v

    def __gt__(self, v):
        """Compare greater than 1 (or -1 if not positive)."""
        return 1 > v if self._positive else -1 > v

    @property
    def ndim(self):
        """Return the dimension of *Identity* class, i.e. *None*."""
        return None

    @property
    def shape(self):
        """Return the shape of *Identity* class, i.e. *None*."""
        return _inftup(None)

    @property
    def T(self):
        """Return the *Identity* class as an operator."""
        return self

    def transpose(self):
        """Return the transpose of the *Identity* class, i.e. itself."""
        return self


class _inftup(tuple):
    """An infinitely long tuple of a value repeated infinitely."""

    def __init__(self, val=None):
        self._val = val

    def __getitem__(self, key):
        if isinstance(key, slice):
            return _inftup(self._val)
        return self._val

    def __len__(self):
        return 0

    def __repr__(self):
        return f"({self._val}, {self._val}, ...)"


################################################
#             DEPRECATED FUNCTIONS
################################################

sdInv = deprecate_function(sdinv, "sdInv", removal_version="1.0.0", error=True)

getSubArray = deprecate_function(
    get_subarray, "getSubArray", removal_version="1.0.0", error=True
)

inv3X3BlockDiagonal = deprecate_function(
    inverse_3x3_block_diagonal,
    "inv3X3BlockDiagonal",
    removal_version="1.0.0",
    error=True,
)

inv2X2BlockDiagonal = deprecate_function(
    inverse_2x2_block_diagonal,
    "inv2X2BlockDiagonal",
    removal_version="1.0.0",
    error=True,
)

makePropertyTensor = deprecate_function(
    make_property_tensor,
    "makePropertyTensor",
    removal_version="1.0.0",
    error=True,
)

invPropertyTensor = deprecate_function(
    inverse_property_tensor,
    "invPropertyTensor",
    removal_version="1.0.0",
    error=True,
)



# cython: embedsignature=True, language_level=3
# cython: linetrace=True
import numpy as np
import cython
cimport numpy as np
import scipy.sparse as sp

def _interp_point_1D(np.ndarray[np.float64_t, ndim=1] x, float xr_i):
    """
        given a point, xr_i, this will find which two integers it lies between.

        :param numpy.ndarray x: Tensor vector of 1st dimension of grid.
        :param float xr_i: Location of a point
        :rtype: int,int,float,float
        :return: index1, index2, portion1, portion2
    """
    cdef IIFF xs
    _get_inds_ws(x,xr_i,&xs)
    return xs.i1,xs.i2,xs.w1,xs.w2

cdef struct IIFF:
    np.int64_t i1,i2
    np.float64_t w1,w2

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
cdef np.int64_t _bisect_left(np.float64_t[:] a, np.float64_t x) nogil:
    cdef np.int64_t lo, hi, mid
    lo = 0
    hi = a.shape[0]
    while lo < hi:
      mid = (lo+hi)//2
      if a[mid] < x: lo = mid+1
      else: hi = mid
    return lo

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
cdef np.int64_t _bisect_right(np.float64_t[:] a, np.float64_t x) nogil:
    cdef np.int64_t lo, hi, mid
    lo = 0
    hi = a.shape[0]
    while lo < hi:
      mid = (lo+hi)//2
      if x < a[mid]: hi = mid
      else: lo = mid+1
    return lo

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
cdef void _get_inds_ws(np.float64_t[:] x, np.float64_t xp, IIFF* out) nogil:
    cdef np.int64_t ind = _bisect_right(x,xp)
    cdef np.int64_t nx = x.shape[0]
    out.i2 = ind
    out.i1 = ind-1
    out.i2 = max(min(out.i2,nx-1),0)
    out.i1 = max(min(out.i1,nx-1),0)
    if(out.i1==out.i2):
        out.w1 = 0.5
    else:
        out.w1 = (x[out.i2]-xp)/(x[out.i2]-x[out.i1])
    out.w2 = 1-out.w1

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
def _interpmat1D(np.ndarray[np.float64_t, ndim=1] locs,
                   np.ndarray[np.float64_t, ndim=1] x):
    cdef int nx = x.size
    cdef IIFF xs
    cdef int npts = locs.shape[0]
    cdef int i

    cdef np.ndarray[np.int64_t,ndim=1] inds = np.empty(npts*2,dtype=np.int64)
    cdef np.ndarray[np.float64_t,ndim=1] vals = np.empty(npts*2,dtype=np.float64)
    for i in range(npts):
        _get_inds_ws(x,locs[i],&xs)

        inds[2*i  ] = xs.i1
        inds[2*i+1] = xs.i2
        vals[2*i  ] = xs.w1
        vals[2*i+1] = xs.w2

    return inds,vals

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
def _interpmat2D(np.ndarray[np.float64_t, ndim=2] locs,
                   np.ndarray[np.float64_t, ndim=1] x,
                   np.ndarray[np.float64_t, ndim=1] y):
    cdef int nx,ny
    nx,ny = len(x),len(y)
    cdef int npts = locs.shape[0]
    cdef int i
    cdef IIFF xs,ys

    cdef np.ndarray[np.int64_t,ndim=2] inds = np.empty((npts*4,2),dtype=np.int64)
    cdef np.ndarray[np.float64_t,ndim=1] vals = np.empty(npts*4,dtype=np.float64)
    for i in range(npts):
        _get_inds_ws(x,locs[i,0],&xs)
        _get_inds_ws(y,locs[i,1],&ys)

        inds[4*i  ,0] = xs.i1
        inds[4*i+1,0] = xs.i1
        inds[4*i+2,0] = xs.i2
        inds[4*i+3,0] = xs.i2
        inds[4*i  ,1] = ys.i1
        inds[4*i+1,1] = ys.i2
        inds[4*i+2,1] = ys.i1
        inds[4*i+3,1] = ys.i2

        vals[4*i  ] = xs.w1*ys.w1
        vals[4*i+1] = xs.w1*ys.w2
        vals[4*i+2] = xs.w2*ys.w1
        vals[4*i+3] = xs.w2*ys.w2

    return inds,vals

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
def _interpmat3D(np.ndarray[np.float64_t, ndim=2] locs,
                 np.ndarray[np.float64_t, ndim=1] x,
                 np.ndarray[np.float64_t, ndim=1] y,
                 np.ndarray[np.float64_t, ndim=1] z):

    cdef int nx,ny,nz
    nx,ny,nz = len(x),len(y),len(z)
    cdef IIFF xs,ys,zs
    cdef int npts = locs.shape[0]
    cdef int i

    cdef np.ndarray[np.int64_t,ndim=2] inds = np.empty((npts*8,3),dtype=np.int64)
    cdef np.ndarray[np.float64_t,ndim=1] vals = np.empty(npts*8,dtype=np.float64)
    for i in range(npts):
        _get_inds_ws(x,locs[i,0],&xs)
        _get_inds_ws(y,locs[i,1],&ys)
        _get_inds_ws(z,locs[i,2],&zs)

        inds[8*i  ,0] = xs.i1
        inds[8*i+1,0] = xs.i1
        inds[8*i+2,0] = xs.i2
        inds[8*i+3,0] = xs.i2
        inds[8*i+4,0] = xs.i1
        inds[8*i+5,0] = xs.i1
        inds[8*i+6,0] = xs.i2
        inds[8*i+7,0] = xs.i2

        inds[8*i  ,1] = ys.i1
        inds[8*i+1,1] = ys.i2
        inds[8*i+2,1] = ys.i1
        inds[8*i+3,1] = ys.i2
        inds[8*i+4,1] = ys.i1
        inds[8*i+5,1] = ys.i2
        inds[8*i+6,1] = ys.i1
        inds[8*i+7,1] = ys.i2

        inds[8*i  ,2] = zs.i1
        inds[8*i+1,2] = zs.i1
        inds[8*i+2,2] = zs.i1
        inds[8*i+3,2] = zs.i1
        inds[8*i+4,2] = zs.i2
        inds[8*i+5,2] = zs.i2
        inds[8*i+6,2] = zs.i2
        inds[8*i+7,2] = zs.i2

        vals[8*i  ] = xs.w1*ys.w1*zs.w1
        vals[8*i+1] = xs.w1*ys.w2*zs.w1
        vals[8*i+2] = xs.w2*ys.w1*zs.w1
        vals[8*i+3] = xs.w2*ys.w2*zs.w1
        vals[8*i+4] = xs.w1*ys.w1*zs.w2
        vals[8*i+5] = xs.w1*ys.w2*zs.w2
        vals[8*i+6] = xs.w2*ys.w1*zs.w2
        vals[8*i+7] = xs.w2*ys.w2*zs.w2

    return inds,vals

@cython.boundscheck(False)
@cython.cdivision(True)
def _tensor_volume_averaging(mesh_in, mesh_out, values=None, output=None):

    cdef np.int32_t[:] i1_in, i1_out, i2_in, i2_out, i3_in, i3_out
    cdef np.float64_t[:] w1, w2, w3
    w1 = np.array([1.0], dtype=np.float64)
    w2 = np.array([1.0], dtype=np.float64)
    w3 = np.array([1.0], dtype=np.float64)
    i1_in = np.array([0], dtype=np.int32)
    i1_out = np.array([0], dtype=np.int32)
    i2_in = np.array([0], dtype=np.int32)
    i2_out = np.array([0], dtype=np.int32)
    i3_in = np.array([0], dtype=np.int32)
    i3_out = np.array([0], dtype=np.int32)
    cdef int dim = mesh_in.dim
    w1, i1_in, i1_out = _volume_avg_weights(mesh_in.nodes_x, mesh_out.nodes_x)
    if dim > 1:
        w2, i2_in, i2_out = _volume_avg_weights(mesh_in.nodes_y, mesh_out.nodes_y)
    if dim > 2:
        w3, i3_in, i3_out = _volume_avg_weights(mesh_in.nodes_z, mesh_out.nodes_z)

    cdef (np.int32_t, np.int32_t, np.int32_t) w_shape = (w1.shape[0], w2.shape[0], w3.shape[0])
    cdef (np.int32_t, np.int32_t, np.int32_t) mesh_in_shape
    cdef (np.int32_t, np.int32_t, np.int32_t) mesh_out_shape

    nCv_in = [len(h) for h in mesh_in.h]
    nCv_out = [len(h) for h in mesh_out.h]
    if dim == 1:
        mesh_in_shape = (nCv_in[0], 1, 1)
        mesh_out_shape = (nCv_out[0], 1, 1)
    elif dim == 2:
        mesh_in_shape = (nCv_in[0], nCv_in[1], 1)
        mesh_out_shape = (nCv_out[0], nCv_out[1], 1)
    elif dim == 3:
        mesh_in_shape = (*nCv_in, )
        mesh_out_shape = (*nCv_out, )

    cdef np.float64_t[::1, :, :] val_in
    cdef np.float64_t[::1, :, :] val_out
    cdef int i1, i2, i3, i1i, i2i, i3i, i1o, i2o, i3o
    cdef np.float64_t w_3, w_32
    cdef np.float64_t[::1, :, :] vol = mesh_out.cell_volumes.reshape(mesh_out_shape, order='F').astype(np.float64)

    if values is not None:
        # If given a values array, do the operation
        val_in = values.reshape(mesh_in_shape, order='F').astype(np.float64)
        if output is None:
            output = np.zeros(mesh_out.n_cells, dtype=np.float64)
        else:
            output = np.require(output, dtype=np.float64, requirements=['A', 'W'])
        v_o = output.reshape(mesh_out_shape, order='F')
        v_o.fill(0)
        val_out = v_o
        for i3 in range(w_shape[2]):
            i3i = i3_in[i3]
            i3o = i3_out[i3]
            w_3 = w3[i3]
            for i2 in range(w_shape[1]):
                i2i = i2_in[i2]
                i2o = i2_out[i2]
                w_32 = w_3*w2[i2]
                for i1 in range(w_shape[0]):
                    i1i = i1_in[i1]
                    i1o = i1_out[i1]
                    val_out[i1o, i2o, i3o] += w_32*w1[i1]*val_in[i1i, i2i, i3i]/vol[i1o, i2o, i3o]
        return output

    # Else, build and return a sparse matrix representing the operation
    i_i = np.empty(w_shape, dtype=np.int32, order='F')
    i_o = np.empty(w_shape, dtype=np.int32, order='F')
    ws = np.empty(w_shape, dtype=np.float64, order='F')
    cdef np.int32_t[::1,:,:] i_in = i_i
    cdef np.int32_t[::1,:,:] i_out = i_o
    cdef np.float64_t[::1, :, :] w = ws
    for i3 in range(w.shape[2]):
        i3i = i3_in[i3]
        i3o = i3_out[i3]
        w_3 = w3[i3]
        for i2 in range(w.shape[1]):
            i2i = i2_in[i2]
            i2o = i2_out[i2]
            w_32 = w_3*w2[i2]
            for i1 in range(w.shape[0]):
                i1i = i1_in[i1]
                i1o = i1_out[i1]
                w[i1, i2, i3] = w_32*w1[i1]/vol[i1o, i2o, i3o]
                i_in[i1, i2, i3] = (i3i*mesh_in_shape[1] + i2i)*mesh_in_shape[0] + i1i
                i_out[i1, i2, i3] = (i3o*mesh_out_shape[1] + i2o)*mesh_out_shape[0] + i1o
    ws = ws.reshape(-1, order='F')
    i_i = i_i.reshape(-1, order='F')
    i_o = i_o.reshape(-1, order='F')
    A = sp.csr_matrix((ws, (i_o, i_i)), shape=(mesh_out.nC, mesh_in.nC))
    return A

@cython.boundscheck(False)
def _volume_avg_weights(np.float64_t[:] x1, np.float64_t[:] x2):
    cdef int n1 = x1.shape[0]
    cdef int n2 = x2.shape[0]
    cdef np.float64_t[:] xs = np.empty(n1 + n2)
    # Fill xs with uniques and truncate
    cdef int i1, i2, i, ii
    i1 = i2 = i = 0
    while i1<n1 or i2<n2:
        if i1<n1 and i2<n2:
            if x1[i1]<x2[i2]:
                xs[i] = x1[i1]
                i1 += 1
            elif x1[i1]>x2[i2]:
                xs[i] = x2[i2]
                i2 += 1
            else:
                xs[i] = x1[i1]
                i1 += 1
                i2 += 1
        elif i1<n1 and i2==n2:
            xs[i] = x1[i1]
            i1 += 1
        elif i2<n2 and i1==n1:
            xs[i] = x2[i2]
            i2 += 1
        i += 1
    cdef int nh = i-1
    hs = np.empty(nh)
    ix1 = np.empty(nh, dtype=np.int32)
    ix2 = np.empty(nh, dtype=np.int32)

    cdef np.float64_t[:] _hs = hs
    cdef np.int32_t[:] _ix1 = ix1
    cdef np.int32_t[:] _ix2 = ix2
    cdef np.float64_t center

    i1 = i2 = ii = 0
    for i in range(nh):
        center = 0.5*(xs[i]+xs[i+1])
        if x2[0] <= center and center <= x2[n2-1]:
            _hs[ii] = xs[i+1]-xs[i]
            while i1<n1-1 and center>=x1[i1]:
                i1 += 1
            while i2<n2-1 and center>=x2[i2]:
                i2 += 1
            _ix1[ii] = min(max(i1-1, 0), n1-1)
            _ix2[ii] = min(max(i2-1, 0), n2-1)
            ii += 1

    hs = hs[:ii]
    ix1 = ix1[:ii]
    ix2 = ix2[:ii]
    return hs, ix1, ix2



from discretize.utils.interpolation_utils import *  # NOQA F401,F403

raise ImportError(
    "Importing from discretize.interputils is deprecated behavoir. Please import "
    "from discretize.utils. This message will be removed in version 1.0.0 of discretize.",
)



"""Module containing unstructured meshes for discretize."""

import numpy as np
import scipy.sparse as sp
from scipy.spatial import KDTree
from discretize.utils import Identity, invert_blocks, spzeros, cross2d
from discretize.base import BaseMesh
from discretize._extensions.simplex_helpers import (
    _build_faces_edges,
    _build_adjacency,
    _directed_search,
    _interp_cc,
)
from discretize.mixins import InterfaceMixins, SimplexMeshIO


class SimplexMesh(BaseMesh, SimplexMeshIO, InterfaceMixins):
    """Class for traingular (2D) and tetrahedral (3D) meshes.

    Simplex is the abstract term for triangular like elements in an arbitrary dimension.
    Simplex meshes are subdivided into trianglular (in 2D) or tetrahedral (in 3D)
    elements. They are capable of representing abstract geometric surfaces, with widely
    variable element sizes.

    Parameters
    ----------
    nodes : (n_nodes, dim) array_like of float
        Defines every node of the mesh.

    simplices : (n_cells, dim+1) array_like of int
        This array defines the connectivity of nodes to form cells. Each element
        indexes into the `nodes` array. Each row defines which nodes make a given cell.
        This array is sorted along each row and then stored on the mesh.

    Notes
    -----
    Only rudimentary checking of the input nodes and simplices is performed, only
    checking for degenerate simplices who have zero volume. There are no checks for
    overlapping cells, or for the quality of the mesh.

    Examples
    --------
    Here we generate a basic 2D triangular mesh, by triangulating a rectangular domain.

    >>> from discretize import SimplexMesh
    >>> import numpy as np
    >>> import matplotlib.pyplot as plt
    >>> import matplotlib.tri as tri

    First we define the nodes of our mesh

    >>> X, Y = np.mgrid[-20:20:21j, -10:10:11j]
    >>> nodes = np.c_[X.reshape(-1), Y.reshape(-1)]

    Then we triangulate the nodes, here we use matplotlib, but you could also use
    scipy's Delaunay, or any other triangular mesh generator. Essentailly we are
    creating every node in the mesh, and a list of triangles/tetrahedrals defining which
    nodes make of each cell.

    >>> triang = tri.Triangulation(nodes[:, 0], nodes[:, 1])
    >>> simplices = triang.triangles

    Finally we can assemble them into a SimplexMesh

    >>> mesh = SimplexMesh(nodes, simplices)
    >>> mesh.plot_grid()
    >>> plt.show()
    """

    _meshType = "simplex"
    _items = {"nodes", "simplices"}

    def __init__(self, nodes, simplices):
        # grab copies of the nodes and simplices for protection
        nodes = np.asarray(nodes)
        simplices = np.asarray(simplices)
        dim = nodes.shape[1]
        if dim not in [3, 2]:
            raise ValueError(
                f"Mesh must be either 2 or 3 dimensions, nodes has {dim} dimension."
            )
        if simplices.shape[1] != dim + 1:
            raise ValueError(
                "simplices second dimension is not compatible with the mesh dimension. "
                f"Saw {simplices.shape[1]}, and expected {dim + 1}."
            )

        self._nodes = nodes.copy()
        self._nodes.setflags(write="false")
        self._simplices = simplices.copy()
        # sort the simplices by node index to simplify further functions...
        self._simplices.sort(axis=1)
        self._simplices.setflags(write="false")

        if self.cell_volumes.min() == 0.0:
            raise ValueError("Triangulation contains degenerate simplices")

        self._number()

    def _number(self):
        items = _build_faces_edges(self.simplices)
        self._simplex_faces = np.array(items[0])
        self._faces = np.array(items[1])
        self._simplex_edges = np.array(items[2])
        self._edges = np.array(items[3])
        self._n_faces = self._faces.shape[0]
        self._n_edges = self._edges.shape[0]
        if self.dim == 3:
            self._face_edges = np.array(items[4])

    @property
    def simplices(self):
        """The node indices for all simplexes of the mesh.

        This array defines the connectivity of the mesh. For each simplex this array is
        sorted by the node index.

        Returns
        -------
        (n_cells, dim + 1) numpy.ndarray of int
        """
        return self._simplices

    @property
    def neighbors(self):
        """
        The adjacancy graph of the mesh.

        This array contains the adjacent cell index for each cell of the mesh. For each
        cell the i'th neighbor is opposite the i'th node, across the i'th face. If a
        cell has no neighbor in a particular direction, then it is listed as having -1.
        This also implies that this is a boundary face.

        Returns
        -------
        (n_cells, dim + 1) numpy.ndarray of int
        """
        if getattr(self, "_neighbors", None) is None:
            self._neighbors = np.array(
                _build_adjacency(self._simplex_faces, self.n_faces)
            )
        return self._neighbors

    @property
    def transform_and_shift(self):
        """
        The barycentric transformation matrix and shift.

        Returns
        -------
        transform : (n_cells, dim, dim) numpy.ndarray
        shift : (n_cells) numpy.ndarray
        """
        if getattr(self, "_transform", None) is None:
            # compute the barycentric transforms
            points = self.nodes
            simplices = self.simplices

            shift = points[self.simplices[:, -1]]

            T = (points[simplices[:, :-1]] - shift[:, None, :]).transpose((0, 2, 1))

            self._transform = invert_blocks(T)
            self._shift = shift
        return self._transform, self._shift

    @property
    def dim(self):  # NOQA D102
        # Documentation inherited from discretize.base.BaseMesh
        return self.nodes.shape[-1]

    @property
    def n_nodes(self):  # NOQA D102
        # Documentation inherited from discretize.base.BaseMesh
        return self._nodes.shape[0]

    @property
    def nodes(self):  # NOQA D102
        # Documentation inherited from discretize.base.BaseMesh
        return self._nodes

    @property
    def n_cells(self):  # NOQA D102
        # Documentation inherited from discretize.base.BaseMesh
        return self.simplices.shape[0]

    @property
    def cell_centers(self):  # NOQA D102
        # Documentation inherited from discretize.base.BaseMesh
        if getattr(self, "_cell_centers", None) is None:
            self._cell_centers = np.mean(self.nodes[self.simplices], axis=1)
        return self._cell_centers

    @property
    def cell_volumes(self):  # NOQA D102
        # Documentation inherited from discretize.base.BaseMesh
        if getattr(self, "_cell_volumes", None) is None:
            simplex_nodes = self._nodes[self.simplices]
            mats = np.pad(simplex_nodes, ((0, 0), (0, 0), (0, 1)), constant_values=1)
            V1 = np.abs(np.linalg.det(mats))
            V1 /= 6 if self.dim == 3 else 2
            self._cell_volumes = V1
        return self._cell_volumes

    @property
    def n_edges(self):  # NOQA D102
        # Documentation inherited from discretize.base.BaseMesh
        return self._n_edges

    @property
    def edges(self):  # NOQA D102
        if getattr(self, "_edge_locs", None) is None:
            self._edge_locs = np.mean(self.nodes[self._edges], axis=1)
        # Documentation inherited from discretize.base.BaseMesh
        return self._edge_locs

    @property
    def edge_tangents(self):  # NOQA D102
        # Documentation inherited from discretize.base.BaseMesh
        if getattr(self, "_edge_tangents", None) is None:
            tangents = np.diff(self.nodes[self._edges], axis=1).squeeze()
            tangents /= np.linalg.norm(tangents, axis=-1)[:, None]
            self._edge_tangents = tangents
        return self._edge_tangents

    @property
    def edge_lengths(self):  # NOQA D102
        # Documentation inherited from discretize.base.BaseMesh
        if getattr(self, "_edge_lengths", None) is None:
            self._edge_lengths = np.linalg.norm(
                np.diff(self.nodes[self._edges], axis=1).squeeze(), axis=-1
            )
        return self._edge_lengths

    @property
    def n_faces(self):  # NOQA D102
        # Documentation inherited from discretize.base.BaseMesh
        return self._n_faces

    @property
    def faces(self):  # NOQA D102
        # Documentation inherited from discretize.base.BaseMesh
        if getattr(self, "_face_locs", None) is None:
            self._face_locs = np.mean(self.nodes[self._faces], axis=1)
        return self._face_locs

    @property
    def face_areas(self):  # NOQA D102
        # Documentation inherited from discretize.base.BaseMesh
        if self.dim == 2:
            return self.edge_lengths
        else:
            if getattr(self, "_face_areas", None) is None:
                face_nodes = self._nodes[self._faces]
                v01 = face_nodes[:, 1] - face_nodes[:, 0]
                v02 = face_nodes[:, 2] - face_nodes[:, 0]
                self._face_areas = np.linalg.norm(np.cross(v01, v02), axis=1) / 2
            return self._face_areas

    @property
    def face_normals(self):  # NOQA D102
        # Documentation inherited from discretize.base.BaseMesh
        if getattr(self, "_face_normals", None) is None:
            if self.dim == 2:
                # Take the normal as being the cross product of edge_tangents
                # and a unit vector in a "3rd" dimension.
                normal = np.c_[self.edge_tangents[:, 1], -self.edge_tangents[:, 0]]
            else:
                # define normal as |01 x 02|
                # therefore clockwise path about the normal is 0->1->2->0
                face_nodes = self._nodes[self._faces]
                v01 = face_nodes[:, 1] - face_nodes[:, 0]
                v02 = face_nodes[:, 2] - face_nodes[:, 0]
                normal = np.cross(v01, v02)
                normal /= np.linalg.norm(normal, axis=1)[:, None]
            self._face_normals = normal
        return self._face_normals

    @property
    def face_divergence(self):  # NOQA D102
        # Documentation inherited from discretize.base.BaseMesh
        if getattr(self, "_face_divergence", None) is None:
            areas = self.face_areas
            normals = self.face_normals

            # approx outward normals (to figure out if face normal is opposite the outward direction)
            test = self.faces[self._simplex_faces] - self.cell_centers[:, None, :]
            dirs = np.einsum("ijk,ijk->ij", normals[self._simplex_faces], test)

            Aijs = areas[self._simplex_faces] / self.cell_volumes[:, None]
            Aijs[dirs < 0] *= -1

            Aijs = Aijs.reshape(-1)
            ind_ptr = (self.dim + 1) * np.arange(self.n_cells + 1)
            col_inds = self._simplex_faces.reshape(-1)
            self._face_divergence = sp.csr_matrix(
                (Aijs, col_inds, ind_ptr), shape=(self.n_cells, self.n_faces)
            )
        return self._face_divergence

    @property
    def nodal_gradient(self):  # NOQA D102
        # Documentation inherited from discretize.base.BaseMesh
        if getattr(self, "_nodal_gradient", None) is None:
            ind_ptr = 2 * np.arange(self.n_edges + 1)
            col_inds = self._edges.reshape(-1)
            Aijs = ((1.0 / self.edge_lengths[:, None]) * [-1, 1]).reshape(-1)
            self._nodal_gradient = sp.csr_matrix(
                (Aijs, col_inds, ind_ptr), shape=(self.n_edges, self.n_nodes)
            )
        return self._nodal_gradient

    @property
    def edge_curl(self):  # NOQA D102
        # Documentation inherited from discretize.base.BaseMesh
        if getattr(self, "_edge_curl", None) is None:
            dim = self.dim
            n_edges = self.n_edges
            if dim == 2:
                face_edges = self._simplex_edges
                face_nodes = self.nodes[self.simplices]
                face_areas = self.cell_volumes
                n_faces = self.n_cells
            else:
                face_edges = self._face_edges
                face_nodes = self.nodes[self._faces]
                face_areas = self.face_areas
                n_faces = self.n_faces

            ind_ptr = 3 * np.arange(n_faces + 1)
            col_inds = face_edges.reshape(-1)

            # edge tangents point from lower node to higher node
            # clockwise about the face normal on face is path from 0 -> 1 -> 2 -> 0
            # so for each face can take the dot product of the path with the edge tangent

            l01 = face_nodes[:, 1] - face_nodes[:, 0]  # path opposite node 2
            l12 = face_nodes[:, 2] - face_nodes[:, 1]  # path opposite node 0
            l20 = face_nodes[:, 0] - face_nodes[:, 2]  # path opposite node 1

            face_path_tangents = self.edge_tangents[face_edges]

            if dim == 2:
                # need to do an adjustment in 2D for the places where the simplices
                # are not oriented counter clockwise about the +z axis
                # cp = np.cross(l01, -l20)
                # cp is a bunch of 1s (where simplices are CCW) and -1s (where simplices are CW)
                # (but we take the sign here to guard against numerical precision)
                cp = np.sign(cross2d(l20, l01))
                face_areas = face_areas * cp
                # don't due *= here

            Aijs = (
                np.c_[
                    np.einsum("ij,ij->i", face_path_tangents[:, 0], l12),
                    np.einsum("ij,ij->i", face_path_tangents[:, 1], l20),
                    np.einsum("ij,ij->i", face_path_tangents[:, 2], l01),
                ]
                / face_areas[:, None]
            ).reshape(-1)

            self._edge_curl = sp.csr_matrix(
                (Aijs, col_inds, ind_ptr), shape=(n_faces, n_edges)
            )
        return self._edge_curl

    def __validate_model(self, model, invert_model=False):
        n_cells = self.n_cells
        dim = self.dim
        # determines the tensor type of the model and reshapes it properly
        model = np.atleast_1d(model)
        n_aniso = ((dim + 1) * dim) // 2
        if model.size == n_aniso * n_cells:
            # model is fully anisotropic
            # reshape it into a stack of dim x dim matrices
            if model.ndim == 1:
                model = model.reshape((-1, n_aniso), order="F")
            vals = model
            if self.dim == 2:
                model = np.stack(
                    [[vals[:, 0], vals[:, 2]], [vals[:, 2], vals[:, 1]]]
                ).transpose((2, 0, 1))
            else:
                model = np.stack(
                    [
                        [vals[:, 0], vals[:, 3], vals[:, 4]],
                        [vals[:, 3], vals[:, 1], vals[:, 5]],
                        [vals[:, 4], vals[:, 5], vals[:, 2]],
                    ]
                ).transpose((2, 0, 1))
        elif model.size == dim * n_cells:
            if model.ndim == 1:
                model = model.reshape((n_cells, dim), order="F")
            model = model.reshape(-1)

        if invert_model:
            if model.ndim == 1:
                model = 1.0 / model
            else:
                model = invert_blocks(model)
        return model

    def __get_inner_product_projection_matrices(
        self, i_type, with_volume=True, return_pointers=True
    ):
        if getattr(self, "_proj_stash", None) is None:
            self._proj_stash = {}
        key = (i_type, with_volume)
        if key not in self._proj_stash:
            dim = self.dim
            n_cells = self.n_cells
            if i_type == "F":
                vecs = self.face_normals
                n_items = self.n_faces
                simplex_items = self._simplex_faces
                if dim == 2:
                    node_items = np.array([[1, 2], [0, 2], [0, 1]])
                else:
                    node_items = np.array([[1, 2, 3], [0, 2, 3], [0, 1, 3], [0, 1, 2]])
            elif i_type == "E":
                vecs = self.edge_tangents
                n_items = self.n_edges
                simplex_items = self._simplex_edges
                if dim == 2:
                    node_items = np.array([[1, 2], [0, 2], [0, 1]])
                elif dim == 3:
                    node_items = np.array([[1, 2, 3], [0, 2, 4], [0, 1, 5], [3, 4, 5]])

            Ps = []
            # Precalc indptr and values for the projection matrix
            P_indptr = np.arange(dim * n_cells + 1)
            ones = np.ones(dim * n_cells)
            if with_volume:
                V = np.sqrt(self.cell_volumes / (dim + 1))

            # precalculate indices for the block diagonal matrix
            d = np.ones(dim, dtype=int)[:, None] * np.arange(dim)
            t = np.arange(n_cells)
            T_col_inds = (d + t[:, None, None] * dim).reshape(-1)
            T_ind_ptr = dim * np.arange(dim * n_cells + 1)

            for i in range(dim + 1):
                # array which selects the items associated with node i of each simplex...
                item_inds = np.take(simplex_items, node_items[i], axis=1)
                P_col_inds = item_inds.reshape(-1)
                P = sp.csr_matrix(
                    (ones, P_col_inds, P_indptr), shape=(dim * n_cells, n_items)
                )

                item_vectors = vecs[item_inds]
                trans_inv = invert_blocks(item_vectors)
                if with_volume:
                    trans_inv *= V[:, None, None]
                T = sp.csr_matrix(
                    (trans_inv.reshape(-1), T_col_inds, T_ind_ptr),
                    shape=(dim * n_cells, dim * n_cells),
                )
                Ps.append(T @ P)

            self._proj_stash[key] = (Ps, T_col_inds, T_ind_ptr)
        Ps, T_col_inds, T_ind_ptr = self._proj_stash[key]
        if return_pointers:
            return Ps, (T_col_inds, T_ind_ptr)
        else:
            return Ps

    def __get_inner_product(self, i_type, model, invert_model):
        Ps, (T_col_inds, T_ind_ptr) = self.__get_inner_product_projection_matrices(
            i_type
        )
        n_cells = self.n_cells
        dim = self.dim
        if model is None:
            Mu = Identity()
        else:
            model = self.__validate_model(model, invert_model)
            if model.size == 1:
                Mu = sp.diags((model,), (0,), shape=(dim * n_cells, dim * n_cells))
            elif model.size == n_cells:
                Mu = sp.diags(np.repeat(model, dim))
            elif model.size == dim * n_cells:
                # diagonally anisotropic model
                Mu = sp.diags(model)
            elif model.size == (dim * dim) * n_cells:
                Mu = sp.csr_matrix(
                    (model.reshape(-1), T_col_inds, T_ind_ptr),
                    shape=(dim * n_cells, dim * n_cells),
                )
            else:
                raise ValueError("Unrecognized size of model vector")

        A = np.sum([P.T @ Mu @ P for P in Ps])
        return A

    def get_face_inner_product(  # NOQA D102
        self,
        model=None,
        invert_model=False,
        invert_matrix=False,
        do_fast=True,
    ):
        # Documentation inherited from discretize.base.BaseMesh
        if invert_matrix:
            raise NotImplementedError(
                "The inverse of the inner product matrix with a tetrahedral mesh is not supported."
            )
        return self.__get_inner_product("F", model, invert_model)

    def get_edge_inner_product(  # NOQA D102
        self,
        model=None,
        invert_model=False,
        invert_matrix=False,
        do_fast=True,
    ):
        # Documentation inherited from discretize.base.BaseMesh
        if invert_matrix:
            raise NotImplementedError(
                "The inverse of the inner product matrix with a tetrahedral mesh is not supported."
            )
        return self.__get_inner_product("E", model, invert_model)

    def __get_inner_product_deriv_func(self, i_type, model):
        Ps, _ = self.__get_inner_product_projection_matrices(i_type)
        dim = self.dim
        n_cells = self.n_cells
        if model.size == 1:
            tensor_type = 0
        elif model.size == n_cells:
            tensor_type = 1
            col_inds = np.repeat(np.arange(n_cells), dim)
            ind_ptr = np.arange(n_cells * dim + 1)
        elif model.size == dim * n_cells:
            tensor_type = 2
            col_inds = np.arange(dim * n_cells).reshape((n_cells, dim), order="F")
            col_inds = col_inds.reshape(-1)
            ind_ptr = np.arange(n_cells * dim + 1)
        elif model.size == (((dim + 1) * dim) // 2) * n_cells:
            tensor_type = 3
            # create a stencil that goes from the model vector ordering
            # into the anisotropy tensor
            if dim == 2:
                stencil = np.array([[0, 2], [2, 1]])
            elif dim == 3:
                stencil = np.array([[0, 3, 4], [3, 1, 5], [4, 5, 2]])
            col_inds = n_cells * stencil + np.arange(n_cells)[:, None, None]
            col_inds = col_inds.reshape(-1)
            ind_ptr = dim * np.arange(n_cells * dim + 1)
        else:
            raise ValueError("Unrecognized size of model vector")
        inv_items = model.size

        if i_type == "F":
            n_items = self.n_faces
        elif i_type == "E":
            n_items = self.n_edges

        def func(v):
            dMdm = spzeros(n_items, inv_items)
            if tensor_type == 0:
                for P in Ps:
                    dMdm = dMdm + sp.csr_matrix(
                        (P.T * (P * v), (range(n_items), np.zeros(n_items))),
                        shape=(n_items, inv_items),
                    )
            elif tensor_type == 1:
                for P in Ps:
                    ys = P @ v
                    dMdm = dMdm + P.T @ sp.csr_matrix(
                        (ys, col_inds, ind_ptr), shape=(n_cells * dim, inv_items)
                    )
            elif tensor_type == 2:
                for P in Ps:
                    ys = P @ v
                    dMdm = dMdm + P.T @ sp.csr_matrix(
                        (ys, col_inds, ind_ptr), shape=(n_cells * dim, inv_items)
                    )
            elif tensor_type == 3:
                for P in Ps:
                    ys = P @ v
                    ys = np.repeat(ys, dim).reshape((-1, dim, dim))
                    ys = ys.transpose((0, 2, 1)).reshape(-1)
                    dMdm = dMdm + P.T @ sp.csr_matrix(
                        (ys, col_inds, ind_ptr), shape=(n_cells * dim, inv_items)
                    )
            return dMdm

        return func

    def get_face_inner_product_deriv(  # NOQA D102
        self, model, do_fast=True, invert_model=False, invert_matrix=False
    ):
        # Documentation inherited from discretize.base.BaseMesh
        if invert_model:
            raise NotImplementedError(
                "Inverted model derivatives are not supported here"
            )
        if invert_matrix:
            raise NotImplementedError("Inverted matrix derivatives are not supported")
        return self.__get_inner_product_deriv_func("F", model)

    def get_edge_inner_product_deriv(  # NOQA D102
        self, model, do_fast=True, invert_model=False, invert_matrix=False
    ):
        # Documentation inherited from discretize.base.BaseMesh
        if invert_model:
            raise NotImplementedError(
                "Inverted model derivatives are not supported here"
            )
        if invert_matrix:
            raise NotImplementedError("Inverted matrix derivatives are not supported")
        return self.__get_inner_product_deriv_func("E", model)

    def _get_edge_surf_int_proj_mats(self, only_boundary=False, with_area=True):
        """Return the projection operators for integrating edges on each face.

        Parameters
        ----------
        only_boundary : bool, optional
            Whether to only operate on the boundary faces or not.
        with_area : bool, optional
            Whether to include the face area.

        Returns
        -------
        list of (3 * n_faces, n_edges) scipy.sparse.csr_matrix
        """
        # edges associated with each face... can just get the indices of the curl...
        face_edges = self._face_edges
        face_areas = self.face_areas
        if only_boundary:
            bf_inds = self.boundary_face_list
            face_edges = face_edges[bf_inds]
            face_areas = face_areas[bf_inds]
            face_normals = self.boundary_face_outward_normals
        else:
            face_normals = self.face_normals

        # face_edges for these are:
        edge_inds = [[1, 2], [0, 2], [0, 1]]

        n_f = face_edges.shape[0]

        ones = np.ones(n_f * 2)
        P_indptr = np.arange(2 * n_f + 1)

        d = np.ones(3, dtype=int)[:, None] * np.arange(2)
        t = np.arange(n_f)
        T_col_inds = (d + t[:, None, None] * 2).reshape(-1)
        T_ind_ptr = 2 * np.arange(3 * n_f + 1)

        Ps = []

        # translate c to fortran ordering
        C2F_col_inds = np.arange(n_f * 3).reshape((-1, 3), order="C").reshape(-1)
        C2F_row_inds = np.arange(n_f * 3).reshape((-1, 3), order="F").reshape(-1)
        C2F = sp.csr_matrix(
            (np.ones(n_f * 3), (C2F_row_inds, C2F_col_inds)), shape=(n_f * 3, n_f * 3)
        )

        for i in range(3):
            # matrix which selects the edges associate with each of the nodes of each boundary face
            node_edges = face_edges[:, edge_inds[i]]
            P = sp.csr_matrix(
                (ones, node_edges.reshape(-1), P_indptr), shape=(2 * n_f, self.n_edges)
            )

            edge_dirs = self.edge_tangents[node_edges]
            t_for = np.concatenate((edge_dirs, face_normals[:, None, :]), axis=1)
            t_inv = invert_blocks(t_for)
            t_inv = t_inv[:, :, :-1] / 3  # n_edges_per_thing

            if with_area:
                t_inv *= face_areas[:, None, None]

            T = C2F @ sp.csr_matrix(
                (t_inv.reshape(-1), T_col_inds, T_ind_ptr),
                shape=(3 * n_f, 2 * n_f),
            )
            Ps.append((T @ P))
        return Ps

    @property
    def cell_centers_tree(self):
        """A KDTree object built from the cell centers.

        Returns
        -------
        scipy.spatial.KDTree
        """
        if getattr(self, "_cc_tree", None) is None:
            self._cc_tree = KDTree(self.cell_centers)
        return self._cc_tree

    def point2index(self, locs):  # NOQA D102
        # Documentation inherited from discretize.base.BaseMesh
        tree = self.cell_centers_tree
        # for each location, find the nearest cell center as an initial guess for
        # the nearest simplex, then use a directed search to further refine
        _, nearest_cc = tree.query(locs)
        nodes = self.nodes
        simplex_nodes = self.simplices
        transform, shift = self.transform_and_shift
        return _directed_search(
            np.atleast_2d(locs),
            np.atleast_1d(nearest_cc),
            nodes,
            simplex_nodes,
            self.neighbors,
            transform,
            shift,
            return_bary=False,
        )

    def get_interpolation_matrix(  # NOQA D102
        self, loc, location_type="cell_centers", zeros_outside=False, **kwargs
    ):
        # Documentation inherited from discretize.base.BaseMesh
        location_type = self._parse_location_type(location_type)
        tree = self.cell_centers_tree
        # for each location, find the nearest cell center as an initial guess for
        # the nearest simplex, then use a directed search to further refine
        loc = np.atleast_2d(loc)
        _, nearest_cc = tree.query(loc)
        nodes = self.nodes
        simplex_nodes = self.simplices
        transform, shift = self.transform_and_shift

        inds, barys = _directed_search(
            loc,
            nearest_cc,
            nodes,
            simplex_nodes,
            self.neighbors,
            transform,
            shift,
            zeros_outside=zeros_outside,
            return_bary=True,
        )

        if zeros_outside:
            barys[inds == -1] = 0.0

        n_loc = len(loc)
        location_type = self._parse_location_type(location_type)
        if location_type == "nodes":
            nodes_per_cell = self.dim + 1
            ind_ptr = nodes_per_cell * np.arange(n_loc + 1)
            col_inds = simplex_nodes[inds].reshape(-1)
            Aij = barys.reshape(-1)
            n_items = self.n_nodes
        elif location_type == "cell_centers":
            # detemine which node each point is closest to.
            which_node = simplex_nodes[inds, np.argmax(barys, axis=-1)]
            # this matrix can be used to lookup which cells a given node touch,
            # which will also be the cells used to interpolate from.
            mat = self.average_node_to_cell.T[which_node].tocsr()
            # this will overwrite the "mat" matrices data to create the interpolation
            _interp_cc(loc, self.cell_centers, mat.data, mat.indices, mat.indptr)
            if zeros_outside:
                e = np.ones(n_loc)
                e[inds == -1] = 0.0
                mat = sp.diags(e, format="csr") @ mat
            return mat
        else:
            component = location_type[-1]
            if component == "x":
                i_dir = 0
            elif component == "y":
                i_dir = 1
            else:
                i_dir = -1
            if location_type[:-2] == "edges":
                # grab the barycentric transforms associated with each simplex:
                ts = transform[inds, :, i_dir]
                ts = np.hstack((ts, -ts.sum(axis=1)[:, None]))

                # edges_x, edges_y, edges_z
                # grab edge lengths
                lengths = self.edge_lengths

                # use 1-form Whitney basis functions for (edges)
                edges = self._simplex_edges[inds]

                # (1, 2), (0, 2), (0, 1)
                e0 = (barys[:, 1] * ts[:, 2] - barys[:, 2] * ts[:, 1]) * lengths[
                    edges[:, 0]
                ]
                e1 = (barys[:, 0] * ts[:, 2] - barys[:, 2] * ts[:, 0]) * lengths[
                    edges[:, 1]
                ]
                e2 = (barys[:, 0] * ts[:, 1] - barys[:, 1] * ts[:, 0]) * lengths[
                    edges[:, 2]
                ]
                if self.dim == 3:
                    # (0, 3), (1, 3), (2, 3)
                    e3 = (barys[:, 0] * ts[:, 3] - barys[:, 3] * ts[:, 0]) * lengths[
                        edges[:, 3]
                    ]
                    e4 = (barys[:, 1] * ts[:, 3] - barys[:, 3] * ts[:, 1]) * lengths[
                        edges[:, 4]
                    ]
                    e5 = (barys[:, 2] * ts[:, 3] - barys[:, 3] * ts[:, 2]) * lengths[
                        edges[:, 5]
                    ]
                    Aij = np.c_[e0, e1, e2, e3, e4, e5].reshape(-1)
                    ind_ptr = 6 * np.arange(n_loc + 1)
                else:
                    Aij = np.c_[e0, e1, e2].reshape(-1)
                    ind_ptr = 3 * np.arange(n_loc + 1)
                col_inds = edges.reshape(-1)
                n_items = self.n_edges
            elif location_type[:-2] == "faces":
                # grab the barycentric transforms associated with each simplex:
                ts = transform[inds, :]
                ts = np.hstack((ts, -ts.sum(axis=1)[:, None]))
                # use  Whitney 2 - form basis functions for face vector interp
                faces = self._simplex_faces[inds]
                areas = self.face_areas

                # [1, 2], [0, 2], [0, 1]
                if self.dim == 2:
                    # i j k
                    # t0 t1 t2
                    # 0 0 1
                    # t1 * i - t0 * j

                    f0 = (
                        cross2d(barys[:, 1:], ts[:, 1:, 1 - i_dir]) * areas[faces[:, 0]]
                    )
                    f1 = (
                        cross2d(barys[:, [0, 2]], ts[:, [0, 2], 1 - i_dir])
                        * areas[faces[:, 1]]
                    )
                    f2 = (
                        cross2d(barys[:, :-1], ts[:, :-1, 1 - i_dir])
                        * areas[faces[:, 2]]
                    )
                    if i_dir == 1:
                        f0 *= -1
                        f1 *= -1
                        f2 *= -1
                    Aij = np.c_[f0, f1, f2].reshape(-1)
                    ind_ptr = 3 * np.arange(n_loc + 1)
                else:
                    # [1, 2, 3], [0, 2, 3], [0, 1, 3], [0, 1, 2]
                    # f123 = (L1 * G_2 x G_3 + L2 * G_3 x G_1 + L3 * G_1 x G_2)
                    # f023 = (L0 * G_2 x G_3 + L2 * G_3 x G_0 + L3 * G_0 x G_2)
                    # f013 = (L0 * G_1 x G_3 + L1 * G_3 x G_0 + L3 * G_0 x G_1)
                    # f012 = (L0 * G_1 x G_2 + L1 * G_2 x G_0 + L2 * G_0 x G_1)
                    f0 = (
                        2
                        * (
                            barys[:, 1] * (np.cross(ts[:, 2], ts[:, 3])[:, i_dir])
                            + barys[:, 2] * (np.cross(ts[:, 3], ts[:, 1])[:, i_dir])
                            + barys[:, 3] * (np.cross(ts[:, 1], ts[:, 2])[:, i_dir])
                        )
                        * areas[faces[:, 0]]
                    )
                    f1 = (
                        2
                        * (
                            barys[:, 0] * (np.cross(ts[:, 2], ts[:, 3])[:, i_dir])
                            + barys[:, 2] * (np.cross(ts[:, 3], ts[:, 0])[:, i_dir])
                            + barys[:, 3] * (np.cross(ts[:, 0], ts[:, 2])[:, i_dir])
                        )
                        * areas[faces[:, 1]]
                    )
                    f2 = (
                        2
                        * (
                            barys[:, 0] * (np.cross(ts[:, 1], ts[:, 3])[:, i_dir])
                            + barys[:, 1] * (np.cross(ts[:, 3], ts[:, 0])[:, i_dir])
                            + barys[:, 3] * (np.cross(ts[:, 0], ts[:, 1])[:, i_dir])
                        )
                        * areas[faces[:, 2]]
                    )
                    f3 = (
                        2
                        * (
                            barys[:, 0] * (np.cross(ts[:, 1], ts[:, 2])[:, i_dir])
                            + barys[:, 1] * (np.cross(ts[:, 2], ts[:, 0])[:, i_dir])
                            + barys[:, 2] * (np.cross(ts[:, 0], ts[:, 1])[:, i_dir])
                        )
                        * areas[faces[:, 3]]
                    )
                    Aij = np.c_[f0, f1, f2, f3].reshape(-1)
                    ind_ptr = 4 * np.arange(n_loc + 1)
                col_inds = faces.reshape(-1)
                n_items = self.n_faces
        return sp.csr_matrix((Aij, col_inds, ind_ptr), shape=(n_loc, n_items))

    @property
    def average_node_to_cell(self):  # NOQA D102
        # Documentation inherited from discretize.base.BaseMesh
        if getattr(self, "_average_node_to_cell", None) is None:
            nodes_per_cell = self.dim + 1
            n_cells = self.n_cells

            ind_ptr = nodes_per_cell * np.arange(n_cells + 1)
            col_inds = self.simplices.reshape(-1)
            Aij = np.full(nodes_per_cell * n_cells, 1 / nodes_per_cell)
            self._average_node_to_cell = sp.csr_matrix(
                (Aij, col_inds, ind_ptr), shape=(n_cells, self.n_nodes)
            )
        return self._average_node_to_cell

    @property
    def average_node_to_face(self):  # NOQA D102
        # Documentation inherited from discretize.base.BaseMesh
        if getattr(self, "_average_node_to_face", None) is None:
            nodes_per_face = self.dim
            n_faces = self.n_faces

            ind_ptr = nodes_per_face * np.arange(n_faces + 1)
            col_inds = self._faces.reshape(-1)
            Aij = np.full(nodes_per_face * n_faces, 1 / nodes_per_face)
            self._average_node_to_face = sp.csr_matrix(
                (Aij, col_inds, ind_ptr), shape=(n_faces, self.n_nodes)
            )
        return self._average_node_to_face

    @property
    def average_node_to_edge(self):  # NOQA D102
        # Documentation inherited from discretize.base.BaseMesh
        if getattr(self, "_average_node_to_edge", None) is None:
            n_edges = self.n_edges

            ind_ptr = 2 * np.arange(n_edges + 1)
            col_inds = self._edges.reshape(-1)
            Aij = np.full(2 * n_edges, 0.5)
            self._average_node_to_edge = sp.csr_matrix(
                (Aij, col_inds, ind_ptr), shape=(n_edges, self.n_nodes)
            )
        return self._average_node_to_edge

    @property
    def average_cell_to_node(self):
        """Averaging matrix from cell centers to nodes.

        The averaging operation uses a volume weighting scheme.

        Returns
        -------
        (n_nodes, n_cells) scipy.sparse.csr_matrix
        """
        # this reproduces linear functions everywhere except on the boundary nodes
        if getattr(self, "_average_cell_to_node", None) is None:
            simps = self.simplices
            cells = np.broadcast_to(
                np.arange(self.n_cells)[:, None], simps.shape
            ).reshape(-1)
            weights = np.broadcast_to(self.cell_volumes[:, None], simps.shape).reshape(
                -1
            )
            simps = simps.reshape(-1)

            A = sp.csr_matrix(
                (weights, (simps, cells)), shape=(self.n_nodes, self.n_cells)
            )
            norm = sp.diags(1.0 / np.asarray(A.sum(axis=1))[:, 0])
            self._average_cell_to_node = norm @ A
        return self._average_cell_to_node

    @property
    def average_cell_to_edge(self):  # NOQA D102
        # Documentation inherited from discretize.base.BaseMesh
        # Simple averaging of all cells with a common edge
        if getattr(self, "_average_cell_to_edge", None) is None:
            simps = self._simplex_edges
            cells = np.broadcast_to(
                np.arange(self.n_cells)[:, None], simps.shape
            ).reshape(-1)
            weights = np.broadcast_to(self.cell_volumes[:, None], simps.shape).reshape(
                -1
            )
            simps = simps.reshape(-1)

            A = sp.csr_matrix(
                (weights, (simps, cells)), shape=(self.n_edges, self.n_cells)
            )
            norm = sp.diags(1.0 / np.asarray(A.sum(axis=1))[:, 0])
            self._average_cell_to_edge = norm @ A
        return self._average_cell_to_edge

    @property
    def average_face_to_cell_vector(self):  # NOQA D102
        # Documentation inherited from discretize.base.BaseMesh
        if getattr(self, "_average_face_to_cell_vector", None) is None:
            dim = self.dim
            n_cells = self.n_cells
            n_faces = self.n_faces

            nodes_per_cell = dim + 1

            Av = sp.csr_matrix((dim * n_cells, n_faces))
            Ps = self.__get_inner_product_projection_matrices(
                "F", with_volume=False, return_pointers=False
            )
            for P in Ps:
                Av = Av + 1 / (nodes_per_cell) * P
            # Av needs to be re-ordered to comply with discretize standard
            ind = np.arange(Av.shape[0]).reshape(n_cells, -1).flatten(order="F")
            P = sp.eye(Av.shape[0], format="csr")[ind]
            self._average_face_to_cell_vector = P @ Av

        return self._average_face_to_cell_vector

    @property
    def average_edge_to_cell_vector(self):  # NOQA D102
        # Documentation inherited from discretize.base.BaseMesh
        if getattr(self, "_average_edge_to_cell_vector", None) is None:
            dim = self.dim
            n_cells = self.n_cells
            n_edges = self.n_edges
            nodes_per_cell = dim + 1

            Av = sp.csr_matrix((dim * n_cells, n_edges))
            # Precalc indptr and values for the projection matrix
            Ps = self.__get_inner_product_projection_matrices(
                "E", with_volume=False, return_pointers=False
            )
            for P in Ps:
                Av = Av + 1 / (nodes_per_cell) * P
            # Av needs to be re-ordered to comply with discretize standard
            ind = np.arange(Av.shape[0]).reshape(n_cells, -1).flatten(order="F")
            P = sp.eye(Av.shape[0], format="csr")[ind]
            self._average_edge_to_cell_vector = P @ Av
        return self._average_edge_to_cell_vector

    @property
    def average_face_to_cell(self):  # NOQA D102
        # Documentation inherited from discretize.base.BaseMesh
        if getattr(self, "_average_face_to_cell", None) is None:
            n_cells = self.n_cells
            n_faces = self.n_faces
            col_inds = self._simplex_faces
            n_face_per_cell = col_inds.shape[1]
            Aij = np.full((n_cells, n_face_per_cell), 1.0 / n_face_per_cell)
            row_ptr = np.arange(n_cells + 1) * (n_face_per_cell)

            self._average_face_to_cell = sp.csr_matrix(
                (Aij.reshape(-1), col_inds.reshape(-1), row_ptr),
                shape=(n_cells, n_faces),
            )
        return self._average_face_to_cell

    @property
    def average_edge_to_cell(self):  # NOQA D102
        # Documentation inherited from discretize.base.BaseMesh
        if getattr(self, "_average_edge_to_cell", None) is None:
            n_cells = self.n_cells
            n_edges = self.n_edges
            col_inds = self._simplex_edges
            n_edge_per_cell = col_inds.shape[1]
            Aij = np.full((n_cells, n_edge_per_cell), 1.0 / (n_edge_per_cell))
            row_ptr = np.arange(n_cells + 1) * (n_edge_per_cell)

            self._average_edge_to_cell = sp.csr_matrix(
                (Aij.reshape(-1), col_inds.reshape(-1), row_ptr),
                shape=(n_cells, n_edges),
            )
        return self._average_edge_to_cell

    @property
    def average_edge_to_face(self):  # NOQA D102
        # Documentation inherited from discretize.base.BaseMesh
        if self.dim == 2:
            # in 2D edges are at the same location as faces
            return sp.eye(self.n_edges, self.n_edges)

        if getattr(self, "_average_edge_to_face", None) is None:
            # in 3D there are three edges per face
            n_faces = self.n_faces
            n_edges = self.n_edges
            face_edges = self._face_edges

            ind_ptr = 3 * np.arange(n_faces + 1)
            col_inds = face_edges.reshape(-1)
            Aijs = np.full(3 * self.n_faces, 1 / 3)
            self._average_edge_to_face = sp.csr_matrix(
                (Aijs, col_inds, ind_ptr), shape=(n_faces, n_edges)
            )
        return self._average_edge_to_face

    @property
    def average_cell_to_face(self):  # NOQA D102
        # Documentation inherited from discretize.base.BaseMesh

        if getattr(self, "_average_cell_to_face", None) is None:
            A = self.average_face_to_cell.T
            row_sum = np.asarray(A.sum(axis=-1))[:, 0]
            row_sum[row_sum == 0.0] = 1.0
            self._average_cell_to_face = sp.diags(1.0 / row_sum) @ A
        return self._average_cell_to_face

    @property
    def stencil_cell_gradient(self):  # NOQA D102
        # Documentation inherited from discretize.base.BaseMesh
        # An operator that differences cells on each side of a face
        # in the direction of the face normal
        if getattr(self, "_stencil_cell_gradient", None) is None:
            tests = self.cell_centers[:, None, :] - self.faces[self._simplex_faces]
            Aij = np.sign(
                np.einsum(
                    "ijk, ijk -> ij", tests, self.face_normals[self._simplex_faces]
                )
            ).reshape(-1)
            ind_ptr = 3 * np.arange(self.n_cells + 1)
            col_inds = self._simplex_faces.reshape(-1)

            self._stencil_cell_gradient = sp.csr_matrix(
                (Aij, col_inds, ind_ptr), shape=(self.n_cells, self.n_faces)
            ).T
        return self._stencil_cell_gradient

    @property
    def boundary_face_list(self):
        """Boolean array of faces that lie on the boundary of the mesh.

        Returns
        -------
        (n_faces) numpy.ndarray of bool
        """
        if getattr(self, "_boundary_face_list", None) is None:
            ind_dir = np.where(self.neighbors == -1)
            self._is_boundary_face = self._simplex_faces[ind_dir]
        return self._is_boundary_face

    @property
    def project_face_to_boundary_face(self):  # NOQA D102
        # Documentation inherited from discretize.base.BaseMesh
        return sp.eye(self.n_faces, format="csr")[self.boundary_face_list]

    @property
    def project_edge_to_boundary_edge(self):  # NOQA D102
        # Documentation inherited from discretize.base.BaseMesh
        if self.dim == 2:
            return self.project_face_to_boundary_face
        bound_edges = np.unique(self._face_edges[self.boundary_face_list])
        return sp.eye(self.n_edges, format="csr")[bound_edges]

    @property
    def project_node_to_boundary_node(self):  # NOQA D102
        # Documentation inherited from discretize.base.BaseMesh
        bound_nodes = np.unique(self._faces[self.boundary_face_list])
        return sp.eye(self.n_nodes, format="csr")[bound_nodes]

    @property
    def boundary_nodes(self):  # NOQA D102
        # Documentation inherited from discretize.base.BaseMesh
        bound_nodes = np.unique(self._faces[self.boundary_face_list])
        return self.nodes[bound_nodes]

    @property
    def boundary_edges(self):  # NOQA D102
        # Documentation inherited from discretize.base.BaseMesh
        if self.dim == 2:
            return self.boundary_faces
        bound_nodes = np.unique(self._face_edges[self.boundary_face_list])
        return self.edges[bound_nodes]

    @property
    def boundary_faces(self):  # NOQA D102
        # Documentation inherited from discretize.base.BaseMesh
        return self.faces[self.boundary_face_list]

    @property
    def boundary_face_outward_normals(self):  # NOQA D102
        # Documentation inherited from discretize.base.BaseMesh
        bound_cells, which_face = np.where(self.neighbors == -1)
        bound_faces = self._simplex_faces[(bound_cells, which_face)]
        bound_face_normals = self.face_normals[bound_faces]

        out_ish = self.faces[bound_faces] - self.cell_centers[bound_cells]
        direc = np.sign(np.einsum("ij,ij->i", bound_face_normals, out_ish))
        boundary_face_outward_normals = direc[:, None] * bound_face_normals

        return boundary_face_outward_normals

    @property
    def boundary_face_scalar_integral(self):  # NOQA D102
        # Documentation inherited from discretize.base.BaseMesh
        P = self.project_face_to_boundary_face

        w_h_dot_normal = np.sum(
            (P @ self.face_normals) * self.boundary_face_outward_normals, axis=-1
        )
        A = sp.diags(self.face_areas) @ P.T @ sp.diags(w_h_dot_normal)
        return A

    @property
    def boundary_node_vector_integral(self):  # NOQA D102
        # Documentation inherited from discretize.base.BaseMesh
        Pn = self.project_node_to_boundary_node
        Pf = self.project_face_to_boundary_face
        n_boundary_nodes = Pn.shape[0]

        dA = self.boundary_face_outward_normals * (Pf @ self.face_areas)[:, None]

        Av = Pf @ self.average_node_to_face @ Pn.T

        u_dot_ds = Av.T @ dA
        diags = u_dot_ds.T
        offsets = n_boundary_nodes * np.arange(self.dim)

        return Pn.T @ sp.diags(
            diags, offsets, shape=(n_boundary_nodes, self.dim * n_boundary_nodes)
        )

    @property
    def boundary_edge_vector_integral(self):  # NOQA D102
        # Documentation inherited from discretize.base.BaseMesh
        boundary_faces = self.boundary_face_list
        if self.dim == 2:
            boundary_face_edges = boundary_faces
            dA = (
                self.boundary_face_outward_normals
                * self.face_areas[boundary_faces][:, None]
            )

            # projection matrices
            # for each edge on boundary faces
            Pe = self.project_edge_to_boundary_edge
            n_boundary_edges, n_edges = Pe.shape
            index = boundary_face_edges
            w_cross_n = cross2d(dA, self.edge_tangents[index])
            M_be = (
                sp.csr_matrix((w_cross_n, (index, index)), shape=(n_edges, n_edges))
                @ Pe.T
            )
        else:
            Ps = self._get_edge_surf_int_proj_mats(only_boundary=True, with_area=True)
            # the cross product of a vector defined on each face with the face outward normal...
            # so V cross n = -n cross V
            cx = sp.diags(self.boundary_face_outward_normals[:, 0])
            cy = sp.diags(self.boundary_face_outward_normals[:, 1])
            cz = sp.diags(self.boundary_face_outward_normals[:, 2])
            # the negative cross mat
            cross_mat = sp.bmat([[None, cz, -cy], [-cz, None, cx], [cy, -cx, None]])

            Pf = self.project_face_to_boundary_face
            Pe = self.project_edge_to_boundary_edge
            Av = (Pf @ self.average_edge_to_face) @ Pe.T
            Av = cross_mat @ sp.block_diag((Av, Av, Av))

            M_be = np.sum(Ps).T @ Av
        return M_be

    def __reduce__(self):
        """Return the class and attributes necessary to reconstruct the mesh."""
        return self.__class__, (
            self.nodes,
            self.simplices,
        )
